from capybara.exceptions import ExpectationNotMet
from capybara.helpers import expects_none, matches_count
from capybara.queries.selector_query import SelectorQuery
from capybara.queries.text_query import TextQuery


class MatchersMixin(object):
    def has_selector(self, *args, **kwargs):
        """
        Checks if a given selector is on the page or a descendant of the current node. ::

            page.has_selector("p#foo")
            page.has_selector("xpath", ".//p[@id='foo']")

        By default it will check if the expression occurs at least once, but a different number can
        be specified. ::

            page.has_selector("p.foo", count=4)

        This will check if the expression occurs exactly 4 times.

        It also accepts all options that :meth:`find_all` accepts, such as ``text`` and
        ``visible``. ::

            page.has_selector("li", text="Horse", visible=True)

        ``has_selector`` can also accept XPath expressions generated by the ``xpath-py`` package::

            from xpath import dsl as x

            page.has_selector("xpath", x.descendant("p"))

        Args:
            *args: Variable length argument list for :class:`SelectorQuery`.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: If the expression exists.
        """

        try:
            self.assert_selector(*args, **kwargs)
            return True
        except ExpectationNotMet:
            return False

    def has_no_selector(self, *args, **kwargs):
        """
        Checks if a given selector is not on the page or a descendant of the current node. Usage is
        identical to :meth:`has_selector`.

        Args:
            *args: Variable length argument list for :class:`SelectorQuery`.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it doesn't exist.
        """

        try:
            self.assert_no_selector(*args, **kwargs)
            return True
        except ExpectationNotMet:
            return False

    def matches_selector(self, *args, **kwargs):
        """
        Checks if the current node matches the given selector.

        Args:
            *args: Variable length argument list for :class:`SelectorQuery`.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it matches.
        """

        try:
            self.assert_matches_selector(*args, **kwargs)
            return True
        except ExpectationNotMet:
            return False

    def not_match_selector(self, *args, **kwargs):
        """
        Checks if the current node does not match the given selector. Usage is identical to
        :meth:`has_selector`.

        Args:
            *args: Variable length argument list for :class:`SelectorQuery`.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it doesn't match.
        """

        try:
            self.assert_not_match_selector(*args, **kwargs)
            return True
        except ExpectationNotMet:
            return False

    def assert_selector(self, *args, **kwargs):
        """
        Asserts that a given selector is on the page or a descendant of the current node. ::

            page.assert_selector("p#foo")

        By default it will check if the expression occurs at least once, but a different number can
        be specified. ::

            page.assert_selector("p.foo", count=4)

        This will check if the expression occurs exactly 4 times. See :meth:`find_all` for other
        available result size options.

        If a ``count`` of 0 is specified, it will behave like :meth:`assert_no_selector`; however,
        use of that method is preferred over this one.

        It also accepts all options that :meth:`find_all` accepts, such as ``text`` and
        ``visible``. ::

            page.assert_selector("li", text="Horse", visible=True)

        ``assert_selector`` can also accept XPath expressions generated by the ``xpath-py``
        package::

            from xpath import dsl as x

            page.assert_selector("xpath", x.descendant("p"))

        Args:
            *args: Variable length argument list for :class:`SelectorQuery`.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            True

        Raises:
            ExpectationNotMet: The given selector did not match.
        """

        query = SelectorQuery(*args, **kwargs)

        @self.synchronize(wait=query.wait)
        def assert_selector():
            result = query.resolve_for(self)

            if not (matches_count(len(result), query.options) and
                    (len(result) > 0 or expects_none(query.options))):
                raise ExpectationNotMet(result.failure_message)

            return True

        return assert_selector()

    def assert_no_selector(self, *args, **kwargs):
        """
        Asserts that a given selector is not on the page or a descendant of the current node. Usage
        is identical to :meth:`assert_selector`.

        Query options such as ``count``, ``minimum``, and ``between`` are considered to be an
        integral part of the selector. This will return True, for example, if a page contains 4
        anchors but the query expects 5::

            page.assert_no_selector("a", minimum=1)  # Found, raises ExpectationNotMet
            page.assert_no_selector("a", count=4)    # Found, raises ExpectationNotMet
            page.assert_no_selector("a", count=5)    # Not Found, returns True

        Args:
            *args: Variable length argument list for :class:`SelectorQuery`.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            True

        Raises:
            ExpectationNotMet: The given selector matched.
        """

        query = SelectorQuery(*args, **kwargs)

        @self.synchronize(wait=query.wait)
        def assert_no_selector():
            result = query.resolve_for(self)

            if matches_count(len(result), query.options) and (
                   len(result) > 0 or expects_none(query.options)):
                raise ExpectationNotMet(result.negative_failure_message)

            return True

        return assert_no_selector()

    refute_selector = assert_no_selector
    """ Alias for :meth:`assert_no_selector`. """

    def assert_matches_selector(self, *args, **kwargs):
        """
        Asserts that the current node matches a given selector. ::

            node.assert_matches_selector("p#foo")
            node.assert_matches_selector("xpath", "//p[@id='foo']")

        It also accepts all options that :meth:`find_all` accepts, such as ``text`` and
        ``visible``. ::

            node.assert_matches_selector("li", text="Horse", visible=True)

        Args:
            *args: Variable length argument list for :class:`SelectorQuery`.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            True

        Raises:
            ExpectationNotMet: If the selector does not match.
        """

        query = SelectorQuery(*args, **kwargs)

        @self.synchronize(wait=query.wait)
        def assert_matches_selector():
            result = query.resolve_for(self.query_scope)

            if self not in result:
                raise ExpectationNotMet("Item does not match the provided selector")

            return True

        return assert_matches_selector()

    def assert_not_match_selector(self, *args, **kwargs):
        """
        Asserts that the current node does not match a given selector. See
        :meth:`assert_matches_selector`.

        Args:
            *args: Variable length argument list for :class:`SelectorQuery`.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            True

        Raises:
            ExpectationNotMet: If the selector matches.
        """

        query = SelectorQuery(*args, **kwargs)

        @self.synchronize(wait=query.wait)
        def assert_not_match_selector():
            result = query.resolve_for(self.query_scope)

            if self in result:
                raise ExpectationNotMet("Item matched the provided selector")

            return True

        return assert_not_match_selector()

    refute_matches_selector = assert_not_match_selector
    """ Alias for :meth:`assert_not_match_selector`. """

    def matches_xpath(self, xpath, **kwargs):
        """
        Checks if the current node matches the given XPath expression.

        Args:
            xpath (str | xpath.expression.Expression): The XPath expression to match against the
                current node.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Return:
            bool: Whether it matches.
        """

        return self.matches_selector("xpath", xpath, **kwargs)

    def not_match_xpath(self, xpath, **kwargs):
        """
        Checks if the current node does not match the given XPath expression.

        Args:
            xpath (str | xpath.expression.Expression): The XPath expression to match against the
                current node.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Return:
            bool: Whether it doesn't match.
        """

        return self.not_match_selector("xpath", xpath, **kwargs)

    def matches_css(self, css, **kwargs):
        """
        Checks if the current node matches the given CSS selector.

        Args:
            css (str): The CSS selector to match against the current node.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Return:
            bool: Whether it matches.
        """

        return self.matches_selector("css", css, **kwargs)

    def not_match_css(self, css, **kwargs):
        """
        Checks if the current node does not match the given CSS selector.

        Args:
            css (str): The CSS selector to match against the current node.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Return:
            bool: Whether it doesn't match.
        """

        return self.not_match_selector("css", css, **kwargs)

    def has_xpath(self, query, **kwargs):
        """
        Checks if a given XPath expression is on the page or a descendant of the current node. ::

            session.has_xpath(".//p[@id='foo']")

        ``has_xpath`` can also accept XPath expressions generated by the ``xpath-py`` package::

            from xpath import dsl as x

            session.has_xpath(x.descendant("p"))

        Args:
            query (str): An XPath expression.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Return:
            bool: If the expression exists.
        """

        return self.has_selector("xpath", query, **kwargs)

    def has_no_xpath(self, path, **kwargs):
        """
        Checks if a given XPath expression is not on the page or a descendant of the current node.
        Usage is identical to :meth:`has_xpath`.

        Args:
            path (str): An XPath expression.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Return:
            bool: If the expression does not exist.
        """

        return self.has_no_selector("xpath", path, **kwargs)

    def has_css(self, path, **kwargs):
        """
        Checks if a given CSS selector is on the page or a descendant of the current node. ::

            page.has_css("p#foo")

        By default it will check if the selector occurs at least once, but a different number can
        be specified. ::

            page.has_css("p#foo", count=4)

        This will check if the selector occurs exactly 4 times.

        It also accepts all options that :meth:`find_all` accepts, such as ``text`` and
        ``visible``. ::

            page.has_css("li", text="Horse", visible=True)

        Args:
            path (str): A CSS selector.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Return:
            bool: If the selector exists.
        """

        return self.has_selector("css", path, **kwargs)

    def has_no_css(self, path, **kwargs):
        """
        Checks if a given CSS selector is not on the page or a descendant of the current node.
        Usage is identical to :meth:`has_css`.

        Args:
            path (str): A CSS Selector.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Return:
            bool: If the selector does not exist.
        """

        return self.has_no_selector("css", path, **kwargs)

    def has_button(self, locator, **kwargs):
        """
        Checks if the page or current node has a button with the given text, value, or id.

        Args:
            locator (str): The text, value, or id of a button to check for.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it exists.
        """

        return self.has_selector("button", locator, **kwargs)

    def has_no_button(self, locator, **kwargs):
        """
        Checks if the page or current node has no button with the given text, value, or id.

        Args:
            locator (str): The text, value, or id of a button to check for.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it doesn't exist.
        """

        return self.has_no_selector("button", locator, **kwargs)

    def has_checked_field(self, locator, **kwargs):
        """
        Checks if the page or current node has a radio button or checkbox with the given label,
        value, or id, that is currently checked.

        Args:
            locator (str): The label, name, or id of a checked field.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it exists.
        """

        kwargs["checked"] = True
        return self.has_selector("field", locator, **kwargs)

    def has_no_checked_field(self, locator, **kwargs):
        """
        Checks if the page or current node has no radio button or checkbox with the given label,
        value, or id that is currently checked.

        Args:
            locator (str): The label, name, or id of a checked field.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it doesn't exist.
        """

        kwargs["checked"] = True
        return self.has_no_selector("field", locator, **kwargs)

    def has_field(self, locator, **kwargs):
        """
        Checks if the page or current node has a form field with the given label, name, or id.

        For text fields and other textual fields, such as textareas and HTML5 email/url/etc. fields,
        it's possible to specify a ``value`` argument to specify the text the field should contain::

            page.has_field("Name", value="Jonas")

        Args:
            locator (str): The label, name, or id of a field to check for.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it exists.
        """

        return self.has_selector("field", locator, **kwargs)

    def has_no_field(self, locator, **kwargs):
        """
        Checks if the page or current node has no form field with the given label, name, or id. See
        :meth:`has_field`.

        Args:
            locator (str): The label, name, or id of a field to check for.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it doesn't exist.
        """

        return self.has_no_selector("field", locator, **kwargs)

    def has_link(self, locator, **kwargs):
        """
        Checks if the page or current node has a link with the given text or id.

        Args:
            locator (str): The text or id of a link to check for.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it exists.
        """

        return self.has_selector("link", locator, **kwargs)

    def has_no_link(self, locator, **kwargs):
        """
        Checks if the page or current node has no link with the given text or id.

        Args:
            locator (str): The text or id of a link to check for.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it doesn't exist.
        """

        return self.has_no_selector("link", locator, **kwargs)

    def has_select(self, locator, **kwargs):
        """
        Checks if the page or current node has a select field with the given label, name, or id.

        It can be specified which option should currently be selected::

            page.has_select("Language", selected="German")

        For multiple select boxes, several options may be specified::

            page.has_select("Language", selected=["English", "German"])

        It's also possible to check if the exact set of options exists for this select box::

            page.has_select("Language", options=["English", "German", "Spanish"])

        You can also check for a partial set of options::

            page.has_select("Language", with_options=["English", "German"])

        Args:
            locator (str): The label, name, or id of a select box.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it exists.
        """

        return self.has_selector("select", locator, **kwargs)

    def has_no_select(self, locator, **kwargs):
        """
        Checks if the page or current node has no select field with the given label, name, or id.
        See :meth:`has_select`.

        Args:
            locator (str): The label, name, or id of a select box.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it doesn't exist.
        """

        return self.has_no_selector("select", locator, **kwargs)

    def has_table(self, locator, **kwargs):
        """
        Checks if the page or current node has a table with the given id or caption::

            page.has_table("People")

        Args:
            locator (str): The id or caption of a table.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it exists.
        """

        return self.has_selector("table", locator, **kwargs)

    def has_no_table(self, locator, **kwargs):
        """
        Checks if the page or current node has no table with the given id or caption. See
        :meth:`has_table`.

        Args:
            locator (str): The id or caption of a table.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it doesn't exist.
        """

        return self.has_no_selector("table", locator, **kwargs)

    def has_unchecked_field(self, locator, **kwargs):
        """
        Checks if the page or current node has a radio button or checkbox with the given label,
        value, or id, that is currently unchecked.

        Args:
            locator (str): The label, name, or id of an unchecked field.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it exists.
        """

        kwargs["checked"] = False
        return self.has_selector("field", locator, **kwargs)

    def has_no_unchecked_field(self, locator, **kwargs):
        """
        Checks if the page or current node has no radio button or checkbox with the given label,
        value, or id, that is currently unchecked.

        Args:
            locator (str): The label, name, or id of an unchecked field.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            bool: Whether it doesn't exist.
        """

        kwargs["checked"] = False
        return self.has_no_selector("field", locator, **kwargs)

    def assert_text(self, *args, **kwargs):
        """
        Asserts that the page or current node has the given text content, ignoring any HTML tags.

        Args:
            *args: Variable length argument list for :class:`TextQuery`.
            **kwargs: Arbitrary keyword arguments for :class:`TextQuery`.

        Returns:
            True

        Raises:
            ExpectationNotMet: If the assertion hasn't succeeded during the wait time.
        """

        query = TextQuery(*args, **kwargs)

        @self.synchronize(wait=query.wait)
        def assert_text():
            count = query.resolve_for(self)

            if not (matches_count(count, query.options) and
                    (count > 0 or expects_none(query.options))):
                raise ExpectationNotMet(query.failure_message)

            return True

        return assert_text()

    def assert_no_text(self, *args, **kwargs):
        """
        Asserts that the page or current node doesn't have the given text content, ignoring any
        HTML tags.

        Args:
            *args: Variable length argument list for :class:`TextQuery`.
            **kwargs: Arbitrary keyword arguments for :class:`TextQuery`.

        Returns:
            True

        Raises:
            ExpectationNotMet: If the assertion hasn't succeeded during the wait time.
        """

        query = TextQuery(*args, **kwargs)

        @self.synchronize(wait=query.wait)
        def assert_no_text():
            count = query.resolve_for(self)

            if matches_count(count, query.options) and (
                   count > 0 or expects_none(query.options)):
                raise ExpectationNotMet(query.negative_failure_message)

            return True

        return assert_no_text()

    def has_text(self, *args, **kwargs):
        """
        Checks if the page or current node has the given text content, ignoring any HTML tags.

        Whitespaces are normalized in both the node's text and the passed text parameter. Note that
        whitespace isn't normalized in a passed regular expression as normalizing whitespace in a
        regular expression isn't easy and doesn't seem to be worth it.

        By default it will check if the text occurs at least once, but a different number can be
        specified. ::

            page.has_text("lorem ipsum", between=range(2, 5))

        Args:
            *args: Variable length argument list for :class:`TextQuery`.
            **kwargs: Arbitrary keyword arguments for :class:`TextQuery`.

        Returns:
            bool: Whether it exists.
        """

        try:
            return self.assert_text(*args, **kwargs)
        except ExpectationNotMet:
            return False

    has_content = has_text
    """ Alias for :meth:`has_text`. """

    def has_no_text(self, *args, **kwargs):
        """
        Checks if the page or current node does not have the given text content, ignoring any HTML
        tags and normalizing whitespace.

        Args:
            *args: Variable length argument list for :class:`TextQuery`.
            **kwargs: Arbitrary keyword arguments for :class:`TextQuery`.

        Returns:
            bool: Whether it doesn't exist.
        """

        try:
            return self.assert_no_text(*args, **kwargs)
        except ExpectationNotMet:
            return False

    has_no_content = has_no_text
    """ Alias for :meth:`has_no_text`. """



import pytest
import re

from capybara.exceptions import ExpectationNotMet


class TestAssertTitle:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_js")

    def test_is_true_if_the_page_title_contains_the_given_string(self, session):
        assert session.assert_title("js") is True

    def test_is_true_when_given_an_empty_string(self, session):
        assert session.assert_title("") is True

    def test_allows_regex_matches(self, session):
        assert session.assert_title(re.compile(r"w[a-z]{3}_js")) is True
        with pytest.raises(ExpectationNotMet) as excinfo:
            session.assert_title(re.compile(r"w[a-z]{10}_js"))
        assert "expected 'with_js' to match 'w[a-z]{10}_js'" in str(excinfo.value)

    @pytest.mark.requires("js")
    def test_waits_for_title(self, session):
        session.click_link("Change title")
        assert session.assert_title("changed title") is True

    def test_raises_error_if_the_page_title_does_not_contain_the_given_string(self, session):
        with pytest.raises(ExpectationNotMet) as excinfo:
            session.assert_title("monkey")
        assert "expected 'with_js' to include 'monkey'" in str(excinfo.value)

    def test_normalizes_the_given_title(self, session):
        session.assert_title("  with_js  ")

    def test_normalizes_given_title_in_error_message(self, session):
        with pytest.raises(ExpectationNotMet) as excinfo:
            session.assert_title(2)
        assert "expected 'with_js' to include '2'" in str(excinfo.value)


class TestAssertNoTitle:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_js")

    def test_raises_error_if_the_title_contains_the_given_string(self, session):
        with pytest.raises(ExpectationNotMet) as excinfo:
            session.assert_no_title("with_j")
        assert "expected 'with_js' not to include 'with_j'" in str(excinfo.value)

    def test_allows_regex_matches(self, session):
        with pytest.raises(ExpectationNotMet) as excinfo:
            session.assert_no_title(re.compile(r"w[a-z]{3}_js"))
        assert "expected 'with_js' not to match 'w[a-z]{3}_js'" in str(excinfo.value)
        session.assert_no_title(re.compile(r"monkey"))

    @pytest.mark.requires("js")
    def test_waits_for_title_to_disappear(self, session):
        session.click_link("Change title")
        assert session.assert_no_title("with_js") is True

    def test_is_true_if_the_title_does_not_contain_the_string(self, session):
        assert session.assert_no_title("monkey") is True



import pytest

from capybara.exceptions import ElementNotFound
from capybara.tests.helpers import extract_results


class TestSelect:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")

    def test_returns_the_value_of_the_first_option(self, session):
        assert session.find_field("Title").value == "Mrs"

    def test_returns_the_value_of_the_selected_option(self, session):
        session.select("Miss", field="Title")
        assert session.find_field("Title").value == "Miss"

    def test_allows_selecting_options_where_they_are_the_only_inexact_match(self, session):
        session.select("Mis", field="Title")
        assert session.find_field("Title").value == "Miss"

    def test_returns_the_value_attribute_rather_than_content_if_present(self, session):
        assert session.find_field("Locale").value == "en"

    def test_selects_an_option_from_a_select_box_by_id(self, session):
        session.select("Finish", field="form_locale")
        session.click_button("awesome")
        assert extract_results(session)["form[locale]"] == "fi"

    def test_selects_an_option_from_a_select_box_by_label(self, session):
        session.select("Finish", field="Locale")
        session.click_button("awesome")
        assert extract_results(session)["form[locale]"] == "fi"

    def test_selects_an_option_without_giving_a_select_box(self, session):
        session.select("Swedish")
        session.click_button("awesome")
        assert extract_results(session)["form[locale]"] == "sv"

    def test_escapes_quotes(self, session):
        session.select("John's made-up language", field="Locale")
        session.click_button("awesome")
        assert extract_results(session)["form[locale]"] == "jo"

    def test_obeys_field(self, session):
        session.select("Miss", field="Other title")
        session.click_button("awesome")
        results = extract_results(session)
        assert results["form[other_title]"] == "Miss"
        assert results["form[title]"] != "Miss"

    def test_matches_labels_with_preceding_or_trailing_whitespace(self, session):
        session.select("Lojban", field="Locale")
        session.click_button("awesome")
        assert extract_results(session)["form[locale]"] == "jbo"

    def test_raises_an_error_with_a_locator_that_does_not_exist(self, session):
        with pytest.raises(ElementNotFound):
            session.select("foo", field="does not exist")

    def test_raises_an_error_with_an_option_that_does_not_exist(self, session):
        with pytest.raises(ElementNotFound):
            session.select("Does not Exist", field="form_locale")

    def test_raises_an_error_for_a_disabled_select(self, session):
        with pytest.raises(ElementNotFound):
            session.select("Should not see me", field="Disabled Select")

    def test_does_not_select_a_disabled_option(self, session):
        with pytest.warns(None) as record:
            session.select("Other", field="form_title")
        assert len(record) == 1
        assert "Attempt to select disabled option" in record[0].message.args[0]
        assert session.find_field("form_title").value != "Other"

    def test_approximately_matches_select_box(self, session):
        session.select("Finish", field="Loc")
        session.click_button("awesome")
        assert extract_results(session)["form[locale]"] == "fi"

    def test_approximately_matches_option(self, session):
        session.select("Fin", field="Locale")
        session.click_button("awesome")
        assert extract_results(session)["form[locale]"] == "fi"

    def test_approximately_matches_option_when_field_not_given(self, session):
        session.select("made-up language")
        session.click_button("awesome")
        assert extract_results(session)["form[locale]"] == "jo"

    def test_does_not_approximately_match_select_box_if_exact_is_true(self, session):
        with pytest.raises(ElementNotFound):
            session.select("Finish", field="Loc", exact=True)

    def test_does_not_approximately_match_option_if_exact_is_true(self, session):
        with pytest.raises(ElementNotFound):
            session.select("Fin", field="Locale", exact=True)

    def test_does_not_approximately_match_option_when_field_not_given_if_exact_is_true(self, session):
        with pytest.raises(ElementNotFound):
            session.select("made-up language", exact=True)


class TestMultipleSelect:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")

    def test_returns_an_empty_value(self, session):
        assert session.find_field("Language").value == []

    def test_returns_the_value_of_the_selected_options(self, session):
        session.select("Ruby", field="Language")
        session.select("Javascript", field="Language")
        languages = session.find_field("Language").value
        assert "Ruby" in languages
        assert "Javascript" in languages

    def test_selects_one_option(self, session):
        session.select("Ruby", field="Language")
        session.click_button("awesome")
        assert extract_results(session).getlist("form[languages][]") == ["Ruby"]

    def test_selects_multiple_options(self, session):
        session.select("Ruby", field="Language")
        session.select("Javascript", field="Language")
        session.click_button("awesome")
        languages = extract_results(session).getlist("form[languages][]")
        assert "Ruby" in languages
        assert "Javascript" in languages

    def test_remains_selected_if_already_selected(self, session):
        session.select("Ruby", field="Language")
        session.select("Javascript", field="Language")
        session.select("Ruby", field="Language")
        session.click_button("awesome")
        languages = extract_results(session).getlist("form[languages][]")
        assert "Ruby" in languages
        assert "Javascript" in languages

    def test_returns_value_attribute_rather_than_content_if_present(self, session):
        assert "thermal" in session.find_field("Underwear").value



import sys

PY2 = sys.version_info[0] == 2

if PY2:
    from urllib import quote, unquote, urlencode
    from urllib2 import URLError, urlopen
    from urlparse import ParseResult, urlparse, parse_qsl

    bytes_ = unicode
    bytes_decode = unicode.encode
    str_encode = str.decode

    def wsgi_decode_body(body):
        return body

    def wsgi_encode_body(body):
        return body
else:
    from urllib.error import URLError
    from urllib.request import urlopen
    from urllib.parse import ParseResult, urlparse, parse_qsl, quote, unquote, urlencode

    bytes_ = bytes
    bytes_decode = bytes.decode
    str_encode = str.encode

    def wsgi_decode_body(body):
        return body.decode()

    def wsgi_encode_body(body):
        return body.encode()



import pytest

from capybara.exceptions import ElementNotFound
from capybara.tests.helpers import extract_results


class TestFillIn:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")

    def test_fills_in_a_text_field_by_id(self, session):
        session.fill_in("form_first_name", value="Harry")
        session.click_button("awesome")
        assert extract_results(session)["form[first_name]"] == "Harry"

    def test_fills_in_a_text_field_by_name(self, session):
        session.fill_in("form[last_name]", value="Green")
        session.click_button("awesome")
        assert extract_results(session)["form[last_name]"] == "Green"

    def test_fills_in_a_text_field_by_label_without_for(self, session):
        session.fill_in("First Name", value="Harry")
        session.click_button("awesome")
        assert extract_results(session)["form[first_name]"] == "Harry"

    def test_fills_in_a_url_field_by_label_without_for(self, session):
        session.fill_in("Html5 Url", value="http://www.avenueq.com")
        session.click_button("html5_submit")
        assert extract_results(session)["form[html5_url]"] == "http://www.avenueq.com"

    def test_fills_in_a_textarea_by_id(self, session):
        session.fill_in("form_description", value="Texty text")
        session.click_button("awesome")
        assert extract_results(session)["form[description]"] == "Texty text"

    def test_fills_in_a_textarea_by_label(self, session):
        session.fill_in("Description", value="Texty text")
        session.click_button("awesome")
        assert extract_results(session)["form[description]"] == "Texty text"

    def test_fills_in_a_textarea_by_name(self, session):
        session.fill_in("form[description]", value="Texty text")
        session.click_button("awesome")
        assert extract_results(session)["form[description]"] == "Texty text"

    def test_fills_in_a_password_field_by_id(self, session):
        session.fill_in("form_password", value="supasikrit")
        session.click_button("awesome")
        assert extract_results(session)["form[password]"] == "supasikrit"

    def test_fills_in_a_password_field_by_label(self, session):
        session.fill_in("Password", value="supasikrit")
        session.click_button("awesome")
        assert extract_results(session)["form[password]"] == "supasikrit"

    def test_fills_in_a_password_field_by_name(self, session):
        session.fill_in("form[password]", value="supasikrit")
        session.click_button("awesome")
        assert extract_results(session)["form[password]"] == "supasikrit"

    def test_handles_html_in_a_textarea(self, session):
        session.fill_in("form_description", value="is <strong>very</strong> secret!")
        session.click_button("awesome")
        assert extract_results(session)["form[description]"] == "is <strong>very</strong> secret!"

    def test_handles_newlines_in_a_textarea(self, session):
        session.fill_in("form_description", value="\nSome text\n")
        session.click_button("awesome")
        assert extract_results(session)["form[description]"] == "\r\nSome text\r\n"

    def test_fills_in_a_field_with_a_custom_type(self, session):
        session.fill_in("Schmooo", value="Schmooo is the game")
        session.click_button("awesome")
        assert extract_results(session)["form[schmooo]"] == "Schmooo is the game"

    def test_fills_in_a_field_without_a_type(self, session):
        session.fill_in("Phone", value="+1 555 7022")
        session.click_button("awesome")
        assert extract_results(session)["form[phone]"] == "+1 555 7022"

    def test_fills_in_a_text_field_respecting_its_maxlength_attribute(self, session):
        session.fill_in("Zipcode", value="52071350")
        session.click_button("awesome")
        assert extract_results(session)["form[zipcode]"] == "52071"

    @pytest.mark.requires("js")
    def test_waits_for_asynchronous_load(self, session):
        session.visit("/with_js")
        session.click_link("Click me")
        session.fill_in("new_field", value="Testing...")

    def test_raises_an_error_with_a_locator_that_does_not_exist(self, session):
        with pytest.raises(ElementNotFound):
            session.fill_in("does not exist", value="Blah blah")

    def test_raises_an_error_for_a_disabled_field(self, session):
        with pytest.raises(ElementNotFound):
            session.fill_in("Disabled Text Field", value="Blah blah")

    def test_fills_in_an_approximately_matched_field(self, session):
        session.fill_in("Explanation", value="Dude")
        session.click_button("awesome")
        assert extract_results(session)["form[name_explanation]"] == "Dude"

    def test_raises_an_error_for_an_approximately_matched_field_when_exact_is_true(self, session):
        with pytest.raises(ElementNotFound):
            session.fill_in("Explanation", value="Dude", exact=True)



import pytest

import capybara
import capybara.dsl


def pytest_runtest_setup(item):
    if item.get_marker("js"):
        capybara.current_driver = capybara.javascript_driver

    driver = item.get_marker("driver")
    if driver:
        assert len(driver.args) == 1, "exactly one driver must be specified"
        capybara.current_driver = driver.args[0]


def pytest_runtest_teardown():
    capybara.reset_sessions()
    capybara.use_default_driver()


@pytest.fixture
def page():
    return capybara.dsl.page



import os.path
import pytest

from capybara.exceptions import ElementNotFound, FileNotFound

from capybara.tests.helpers import extract_results


class AttachFileTestCase:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")

    @pytest.fixture
    def file_path(self, fixture_path):
        return fixture_path("file.txt")

    @pytest.fixture
    def another_file_path(self, fixture_path):
        return fixture_path("another_file.txt")

    @pytest.fixture
    def jpg_file_path(self, fixture_path):
        return fixture_path("capybara.jpg")


class TestAttachFile(AttachFileTestCase):
    def test_raises_an_error_with_a_locator_that_does_not_exist(self, session, file_path):
        with pytest.raises(ElementNotFound) as excinfo:
            session.attach_file("does not exist", file_path)
        assert "Unable to find file field 'does not exist'" in str(excinfo.value)

    def test_raises_an_error_with_a_path_that_does_not_exist(self, session):
        with pytest.raises(FileNotFound):
            session.attach_file("Image", "/no_such_file.png")

    def test_sets_a_file_path_by_partial_label(self, session, file_path):
        session.attach_file("Imag", file_path)
        session.click_button("awesome")
        assert extract_results(session)["form[image]"] == os.path.basename(file_path)

    def test_does_not_set_a_file_path_by_partial_label_when_exact_is_true(self, session, file_path):
        with pytest.raises(ElementNotFound):
            session.attach_file("Imag", file_path, exact=True)


class TestAttachFileWithNormalForm(AttachFileTestCase):
    def test_sets_a_file_path_by_id(self, session, file_path):
        session.attach_file("form_image", file_path)
        session.click_button("awesome")
        assert extract_results(session)["form[image]"] == os.path.basename(file_path)

    def test_sets_a_file_path_by_label(self, session, file_path):
        session.attach_file("Image", file_path)
        session.click_button("awesome")
        assert extract_results(session)["form[image]"] == os.path.basename(file_path)


class TestAttachFileWithMultipartForm(AttachFileTestCase):
    def test_sets_a_file_path_by_id(self, session, file_path):
        session.attach_file("form_document", file_path)
        session.click_button("Upload Single")
        assert session.has_text(open(file_path).read())

    def test_sets_a_file_path_by_label(self, session, file_path):
        session.attach_file("Single Document", file_path)
        session.click_button("Upload Single")
        assert session.has_text(open(file_path).read())

    def test_does_not_break_if_no_file_is_submitted(self, session):
        session.click_button("Upload Single")
        assert session.has_text("No file uploaded")

    def test_sends_content_type_text_plain_when_uploading_a_text_file(self, session, file_path):
        session.attach_file("Single Document", file_path)
        session.click_button("Upload Single")
        assert session.has_text("text/plain")

    def test_sends_content_type_image_jpeg_when_uploading_an_image(self, session, jpg_file_path):
        session.attach_file("Single Document", jpg_file_path)
        session.click_button("Upload Single")
        assert session.has_text("image/jpeg")

    def test_does_not_break_when_using_html5_multiple_file_input(self, session, file_path):
        session.attach_file("Multiple Documents", file_path)
        session.click_button("Upload Multiple")
        assert session.has_text(open(file_path).read())

    def test_does_not_send_anything_when_attaching_no_files_to_a_multiple_upload_field(self, session):
        session.click_button("Upload Empty Multiple")
        assert session.has_text("Successfully ignored empty file field")



import pytest
import re

import capybara
from capybara.exceptions import ElementNotFound


class TestAssertSelector:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")

    def test_does_not_raise_if_the_given_selector_is_on_the_page(self, session):
        session.assert_selector("xpath", "//p")
        session.assert_selector("css", "p a#foo")
        session.assert_selector("//p[contains(.,'est')]")

    def test_raises_if_the_given_selector_is_not_on_the_page(self, session):
        with pytest.raises(ElementNotFound):
            session.assert_selector("xpath", "//abbr")
        with pytest.raises(ElementNotFound):
            session.assert_selector("css", "p a#doesnotexist")
        with pytest.raises(ElementNotFound):
            session.assert_selector("//p[contains(.,'thisstringisnotonpage')]")

    def test_uses_default_selector(self, session):
        capybara.default_selector = "css"
        with pytest.raises(ElementNotFound):
            session.assert_selector("p a#doesnotexist")
        session.assert_selector("p a#foo")

    def test_respects_scopes(self, session):
        with session.scope("//p[@id='first']"):
            session.assert_selector(".//a[@id='foo']")
            with pytest.raises(ElementNotFound):
                session.assert_selector(".//a[@id='red']")

    def test_is_true_if_the_content_is_on_the_page_the_given_number_of_times(self, session):
        session.assert_selector("//p", count=3)
        session.assert_selector("//p//a[@id='foo']", count=1)
        session.assert_selector("//p[contains(.,'est')]", count=1)

    def test_raises_if_the_content_is_on_the_page_the_given_number_of_times(self, session):
        with pytest.raises(ElementNotFound):
            session.assert_selector("//p", count=6)
        with pytest.raises(ElementNotFound):
            session.assert_selector("//p//a[@id='foo']", count=2)
        with pytest.raises(ElementNotFound):
            session.assert_selector("//p[contains(.,'est')]", count=5)

    def test_raises_if_the_content_is_not_on_the_page_at_all(self, session):
        with pytest.raises(ElementNotFound):
            session.assert_selector("//abbr", count=2)
        with pytest.raises(ElementNotFound):
            session.assert_selector("//p//a[@id='doesnotexist']", count=1)

    def test_discards_all_matches_where_the_given_string_is_not_contained(self, session):
        session.assert_selector("//p//a", text="Redirect", count=1)
        with pytest.raises(ElementNotFound):
            session.assert_selector("//p", text="Doesnotexist")

    def test_discards_all_matches_where_the_given_regex_is_not_matched(self, session):
        session.assert_selector("//p//a", text=re.compile("re[dab]i", re.IGNORECASE), count=1)
        with pytest.raises(ElementNotFound):
            session.assert_selector("//p//a", text=re.compile("Red$"))

    @pytest.mark.requires("js")
    def test_finds_element_if_it_appears_before_given_wait_duration(self, session):
        with capybara.using_wait_time(0.1):
            session.visit("/with_js")
            session.click_link("Click me")
            session.assert_selector("css", "a#has-been-clicked", text="Has been clicked", wait=0.9)


class TestAssertNoSelector:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")

    def test_raises_an_error_if_the_given_selector_is_on_the_page(self, session):
        with pytest.raises(ElementNotFound):
            session.assert_no_selector("xpath", "//p")
        with pytest.raises(ElementNotFound):
            session.assert_no_selector("css", "p a#foo")
        with pytest.raises(ElementNotFound):
            session.assert_no_selector("//p[contains(.,'est')]")

    def test_is_true_if_the_given_selector_is_not_on_the_page(self, session):
        session.assert_no_selector("xpath", "//abbr")
        session.assert_no_selector("css", "p a#doesnotexist")
        session.assert_no_selector("//p[contains(.,'thisstringisnotonpage')]")

    def test_uses_the_default_selector(self, session):
        capybara.default_selector = "css"
        session.assert_no_selector("p a#doesnotexist")
        with pytest.raises(ElementNotFound):
            session.assert_no_selector("p a#foo")

    def test_respects_scopes(self, session):
        with session.scope("//p[@id='first']"):
            with pytest.raises(ElementNotFound):
                session.assert_no_selector(".//a[@id='foo']")
            session.assert_no_selector(".//a[@id='red']")

    def test_raises_an_error_if_the_content_is_on_the_page_the_given_number_of_times(self, session):
        with pytest.raises(ElementNotFound):
            session.assert_no_selector("//p", count=3)
        with pytest.raises(ElementNotFound):
            session.assert_no_selector("//p//a[@id='foo']", count=1)
        with pytest.raises(ElementNotFound):
            session.assert_no_selector("//p[contains(.,'est')]", count=1)

    def test_is_true_if_the_content_is_on_the_page_the_wrong_number_of_times(self, session):
        session.assert_no_selector("//p", count=6)
        session.assert_no_selector("//p//a[@id='foo']", count=2)
        session.assert_no_selector("//p[contains(.,'est')]", count=5)

    def test_is_true_if_the_content_is_not_on_the_page_at_all(self, session):
        session.assert_no_selector("//abbr", count=2)
        session.assert_no_selector("//p//a[@id='doesnotexist']", count=1)

    def test_discards_all_matches_where_the_given_string_is_contained(self, session):
        with pytest.raises(ElementNotFound):
            session.assert_no_selector("//p//a", text="Redirect", count=1)
        session.assert_no_selector("//p", text="Doesnotexist")

    def test_discards_all_matches_where_the_given_regex_is_matched(self, session):
        with pytest.raises(ElementNotFound):
            session.assert_no_selector("//p//a", text=re.compile(r"re[dab]i", re.IGNORECASE), count=1)
        session.assert_no_selector("//p//a", text=re.compile(r"Red$"))

    @pytest.mark.requires("js")
    def test_does_not_find_element_if_it_appears_after_given_wait_duration(self, session):
        session.visit("/with_js")
        session.click_link("Click me")
        session.assert_no_selector("css", "a#has-been-clicked", text="Has been clicked", wait=0.1)



# coding=utf-8
from __future__ import unicode_literals

import pytest
import re

import capybara


class TestHasText:
    def test_is_true_if_the_given_text_is_on_the_page_at_least_once(self, session):
        session.visit("/with_html")
        assert session.has_text("est")
        assert session.has_text("Lorem")
        assert session.has_text("Redirect")

    def test_ignores_tags(self, session):
        session.visit("/with_html")
        assert not session.has_text("""exercitation <a href="/foo">ullamco</a> laboris""")
        assert session.has_text("exercitation ullamco laboris")

    def test_ignores_extra_whitespace_and_newlines(self, session):
        session.visit("/with_html")
        assert session.has_text("text with whitespace")

    def test_ignores_white_space_and_newlines_in_the_search_string(self, session):
        session.visit("/with_html")
        assert session.has_text("text     with \n\n whitespace")

    def test_is_false_if_the_given_text_is_not_on_the_page(self, session):
        session.visit("/with_html")
        assert not session.has_text("xxxxyzzz")
        assert not session.has_text("monkey")

    def test_is_true_if_the_given_unicode_text_is_on_the_page(self, session):
        session.visit("/with_html")
        assert session.has_text("이름")

    def test_is_false_if_the_given_unicode_text_is_not_on_the_page(self, session):
        session.visit("/with_html")
        assert not session.has_text("论坛")

    def test_handles_single_quotes_in_the_text(self, session):
        session.visit("/with-quotes")
        assert session.has_text("can't")

    def test_handles_double_quotes_in_the_text(self, session):
        session.visit("/with-quotes")
        assert session.has_text("\"No,\" he said")

    def test_handles_mixed_single_and_double_quotes_in_the_text(self, session):
        session.visit("/with-quotes")
        assert session.has_text("\"you can't do that.\"")

    def test_is_false_if_text_is_in_the_title_tag_in_the_head(self, session):
        session.visit("/with_js")
        assert not session.has_text("with_js")

    def test_is_false_if_text_is_inside_a_script_tag_in_the_body(self, session):
        session.visit("/with_js")
        assert not session.has_text("a javascript comment")
        assert not session.has_text("aVar")

    def test_is_false_if_the_given_text_is_on_the_page_but_not_visible(self, session):
        session.visit("/with_html")
        assert not session.has_text("Inside element with hidden ancestor")

    def test_is_true_if_all_given_and_text_is_invisible(self, session):
        session.visit("/with_html")
        assert session.has_text("all", "Some of this text is hidden!")

    def test_is_true_if_capybara_ignore_hidden_elements_is_false_and_text_is_invisible(self, session):
        capybara.ignore_hidden_elements = False
        session.visit("/with_html")
        assert session.has_text("Some of this text is hidden!")

    def test_is_true_if_the_text_in_the_page_matches_given_regex(self, session):
        session.visit("/with_html")
        assert session.has_text(re.compile(r"Lorem"))

    def test_is_false_if_the_text_in_the_page_does_not_match_given_regex(self, session):
        session.visit("/with_html")
        assert not session.has_text(re.compile(r"xxxxyzzz"))

    def test_escapes_any_characters_that_would_have_special_meaning_in_a_regex(self, session):
        session.visit("/with_html")
        assert not session.has_text(".orem")

    def test_accepts_non_string_parameters(self, session):
        session.visit("/with_html")
        assert session.has_text(42)

    def test_is_true_when_passed_none(self, session):
        session.visit("/with_html")
        assert session.has_text(None)

    @pytest.mark.requires("js")
    def test_waits_for_text_to_appear(self, session):
        session.visit("/with_js")
        session.click_link("Click me")
        assert session.has_text("Has been clicked")

    def test_is_true_if_the_text_occurs_within_the_range_given(self, session):
        session.visit("/with_count")
        assert session.has_text("count", between=range(1, 4))
        assert session.has_text(re.compile(r"count"), between=range(2, 3))

    def test_is_false_if_the_text_occurs_more_or_fewer_times_than_range(self, session):
        session.visit("/with_count")
        assert not session.has_text("count", between=range(0, 2))
        assert not session.has_text("count", between=range(3, 11))
        assert not session.has_text(re.compile(r"count"), between=range(2, 2))

    def test_is_true_if_the_text_occurs_the_given_number_of_times(self, session):
        session.visit("/with_count")
        assert session.has_text("count", count=2)

    def test_is_false_if_the_text_occurs_a_different_number_of_times_than_given(self, session):
        session.visit("/with_count")
        assert not session.has_text("count", count=0)
        assert not session.has_text("count", count=1)
        assert not session.has_text(re.compile(r"count"), count=3)

    def test_coerces_count_to_an_integer(self, session):
        session.visit("/with_count")
        assert session.has_text("count", count="2")
        assert not session.has_text("count", count="3")

    def test_is_true_when_text_occurs_same_or_fewer_times_than_given(self, session):
        session.visit("/with_count")
        assert session.has_text("count", maximum=2)
        assert session.has_text("count", maximum=3)

    def test_is_false_when_text_occurs_more_times_than_given(self, session):
        session.visit("/with_count")
        assert not session.has_text("count", maximum=1)
        assert not session.has_text(re.compile(r"count"), maximum=0)

    def test_coerces_maximum_to_an_integer(self, session):
        session.visit("/with_count")
        assert session.has_text("count", maximum="2")
        assert not session.has_text("count", maximum="1")

    def test_is_true_when_text_occurs_same_or_more_times_than_given(self, session):
        session.visit("/with_count")
        assert session.has_text("count", minimum=2)
        assert session.has_text(re.compile(r"count"), minimum=0)

    def test_is_false_when_text_occurs_fewer_times_than_given(self, session):
        session.visit("/with_count")
        assert not session.has_text("count", minimum=3)

    def test_coerces_minimum_to_an_integer(self, session):
        session.visit("/with_count")
        assert session.has_text("count", minimum="2")
        assert not session.has_text("count", minimum="3")

    @pytest.mark.requires("js")
    def test_finds_element_if_it_appears_before_given_wait_duration(self, session):
        with capybara.using_wait_time(0.1):
            session.visit("/with_js")
            session.click_link("Click me")
            assert session.has_text("Has been clicked", wait=0.9)


class TestHasNoText:
    def test_is_false_if_the_given_text_is_on_the_page_at_least_once(self, session):
        session.visit("/with_html")
        assert not session.has_no_text("est")
        assert not session.has_no_text("Lorem")
        assert not session.has_no_text("Redirect")

    def test_is_false_if_scoped_to_an_element_which_has_the_text(self, session):
        session.visit("/with_html")
        with session.scope("//a[@title='awesome title']"):
            assert not session.has_no_text("labore")

    def test_is_true_if_scoped_to_an_element_which_does_not_have_the_text(self, session):
        session.visit("/with_html")
        with session.scope("//a[@title='awesome title']"):
            assert session.has_no_text("monkey")

    def test_ignores_tags(self, session):
        session.visit("/with_html")
        assert session.has_no_text("""exercitation <a href="/foo" id="foo">ullamco</a> laboris""")
        assert not session.has_no_text("exercitation ullamco laboris")

    def test_is_true_if_the_given_text_is_not_on_the_page(self, session):
        session.visit("/with_html")
        assert session.has_no_text("xxxxyzzz")
        assert session.has_no_text("monkey")

    def test_handles_single_quotes_in_the_text(self, session):
        session.visit("/with-quotes")
        assert not session.has_no_text("can't")

    def test_handles_double_quotes_in_the_text(self, session):
        session.visit("/with-quotes")
        assert not session.has_no_text("\"you can't do that.\"")

    def test_is_true_if_text_is_in_the_title_tag_in_the_head(self, session):
        session.visit("/with_js")
        assert session.has_no_text("with_js")

    def test_is_true_if_text_is_inside_a_script_tag_in_the_body(self, session):
        session.visit("/with_js")
        assert session.has_no_text("a javascript comment")
        assert session.has_no_text("aVar")

    def test_is_true_if_the_given_text_is_on_the_page_but_not_visible(self, session):
        session.visit("/with_html")
        assert session.has_no_text("Inside element with hidden ancestor")

    def test_is_false_if_all_given_and_text_is_invisible(self, session):
        session.visit("/with_html")
        assert not session.has_no_text("all", "Some of this text is hidden!")

    def test_is_false_if_capybara_ignore_hidden_elements_is_false_and_text_is_invisible(self, session):
        capybara.ignore_hidden_elements = False
        session.visit("/with_html")
        assert not session.has_no_text("Some of this text is hidden!")

    def test_is_true_if_the_text_in_the_page_does_not_match_given_regex(self, session):
        session.visit("/with_html")
        assert session.has_no_text(re.compile(r"xxxxyzzz"))

    def test_is_false_if_the_text_in_the_page_matches_given_regex(self, session):
        session.visit("/with_html")
        assert not session.has_no_text(re.compile(r"Lorem"))

    def test_escapes_any_characters_that_would_have_special_meaning_in_a_regex(self, session):
        session.visit("/with_html")
        assert session.has_no_text(".orem")

    @pytest.mark.requires("js")
    def test_waits_for_text_to_disappear(self, session):
        session.visit("/with_js")
        session.click_link("Click me")
        assert session.has_no_text("I changed it")

    @pytest.mark.requires("js")
    def test_does_not_find_element_if_it_appears_after_given_wait_duration(self, session):
        session.visit("/with_js")
        session.click_link("Click me")
        assert session.has_no_text("Has been clicked", wait=0.1)



import re
from xpath import dsl as x
from xpath.renderer import to_xpath

from capybara.node.document_matchers import DocumentMatchersMixin
from capybara.node.finders import FindersMixin
from capybara.node.matchers import MatchersMixin
from capybara.utils import inner_content


class Simple(FindersMixin, MatchersMixin, DocumentMatchersMixin, object):
    """
    A :class:`Simple` is a simpler version of :class:`Base` which includes only
    :class:`FindersMixin` and :class:`MatchersMixin` and does not include :class:`ActionsMixin`.
    This type of node is returned when using :func:`capybara.string`.

    It is useful in that it does not require a session, an application, or a driver, but can still
    use Capybara's finders and matchers on any string that contains HTML.
    """

    def __init__(self, native):
        if isinstance(native, str):
            from lxml import etree
            native = etree.HTML(native)
        self.native = native

    def __getitem__(self, name):
        """
        Return the given attribute. ::

            element["title"]  # => HTML title attribute

        Args:
            name (str): The attribute name to retrieve.

        Returns:
            str: The value of the attribute.
        """

        return self.native.get(name)

    @property
    def tag_name(self):
        """ str: The tag name of the element. """
        return self.native.tag

    @property
    def path(self):
        """ str: An XPath expression describing where on the page the element can be found. """
        return self.native.getroottree().getpath(self.native)

    @property
    def text(self):
        """ str: The text of the element. """
        return self.native.text

    @property
    def value(self):
        """ str: The value of the form element. """

        if self.tag_name == "textarea":
            return inner_content(self.native)
        elif self.tag_name == "select":
            if self["multiple"] == "multiple":
                selected_options = self._find_xpath(".//option[@selected='selected']")
                return [_get_option_value(option) for option in selected_options]
            else:
                options = (
                    self._find_xpath(".//option[@selected='selected']") +
                    self._find_xpath(".//option"))
                return _get_option_value(options[0]) if options else None
        elif self.tag_name == "input" and self["type"] in ["checkbox", "radio"]:
            return self["value"] or "on"
        else:
            return self["value"]

    @property
    def visible(self):
        """ bool: Whether or not the element is visible. """
        return self._visible()

    @property
    def checked(self):
        """ bool: Whether or not the element is checked. """
        return "checked" in self.native.attrib

    @property
    def disabled(self):
        """ bool: Whether or not the element is disabled. """
        return "disabled" in self.native.attrib

    @property
    def readonly(self):
        """ bool: Whether or not the element is readonly. """
        return "readonly" in self.native.attrib

    @property
    def multiple(self):
        """ bool: Whether or not the select supports multiple values. """
        return "multiple" in self.native.attrib

    @property
    def selected(self):
        """ bool: Whether or not the element is selected. """
        return "selected" in self.native.attrib

    @property
    def allow_reload(self):
        return False

    @allow_reload.setter
    def allow_reload(self, value):
        pass

    @property
    def title(self):
        """ str: The current page title. """
        elements = self.native.xpath("/html/head/title | /html/title")
        return elements[0].text if elements else None

    def synchronize(self, func=None, **kwargs):
        # Simple nodes don't need to wait.
        return func if func else lambda func: func

    def _visible(self, check_ancestor_visibility=True):
        if self.tag_name == "input" and self.native.get("type") == "hidden":
            return False

        if check_ancestor_visibility:
            return not self.native.xpath(
                "./ancestor-or-self::*["
                "contains(@style, 'display:none') or "
                "contains(@style, 'display: none') or "
                "@hidden or "
                "name()='script' or "
                "name()='head'"
                "]")

        if "hidden" in self.native.attrib:
            return False
        if re.compile(r"display:\s?none").search(self.native.get("style", "")):
            return False
        if self.tag_name in ["script", "head"]:
            return False

        return True

    def _find_xpath(self, xpath):
        return self.native.xpath(xpath)

    def _find_css(self, css):
        xpath = to_xpath(x.css(css))
        return self._find_xpath(xpath)


def _get_option_value(option):
    return option.get("value") or inner_content(option)



from selenium.webdriver.common.action_chains import ActionChains

from capybara.exceptions import ReadOnlyElementError, UnselectNotAllowed
from capybara.driver.node import Node as Base
from capybara.helpers import normalize_whitespace


class Node(Base):
    @property
    def tag_name(self):
        return self.native.tag_name

    @property
    def visible(self):
        return self.native.is_displayed()

    @property
    def value(self):
        if self.tag_name == "select" and self.multiple:
            options = self.native.find_elements_by_xpath(".//option")
            selected_options = filter(lambda opt: opt.is_selected(), options)
            return [opt.get_attribute("value") for opt in selected_options]
        else:
            return self.native.get_attribute("value")

    @property
    def selected(self):
        return self.native.is_selected()

    @property
    def multiple(self):
        val = self["multiple"]
        return bool(val) and val != "false"

    @property
    def path(self):
        path = [self] + list(reversed(self._find_xpath("ancestor::*")))

        result = []
        while path:
            node = path.pop(0)
            parent = path[0] if path else None

            if parent:
                siblings = parent._find_xpath(node.tag_name)
                if len(siblings) == 1:
                    result.insert(0, node.tag_name)
                else:
                    index = siblings.index(node)
                    result.insert(0, "{tag}[{i}]".format(tag=node.tag_name, i=index + 1))
            else:
                result.insert(0, node.tag_name)

        return "/" + "/".join(result)

    @property
    def all_text(self):
        text = self.driver.browser.execute_script("return arguments[0].textContent", self.native)
        return normalize_whitespace(text)

    @property
    def visible_text(self):
        return normalize_whitespace(self.native.text)

    def __getitem__(self, name):
        return self.native.get_attribute(name)

    def _find_css(self, css):
        cls = type(self)
        return [cls(self.driver, element)
                for element in self.native.find_elements_by_css_selector(css)]

    def _find_xpath(self, xpath):
        cls = type(self)
        return [cls(self.driver, element)
                for element in self.native.find_elements_by_xpath(xpath)]

    def click(self):
        self.native.click()

    def double_click(self):
        ActionChains(self.driver.browser).double_click(self.native).perform()

    def drag_to(self, element):
        ActionChains(self.driver.browser).drag_and_drop(self.native, element.native).perform()

    def hover(self):
        ActionChains(self.driver.browser).move_to_element(self.native).perform()

    def right_click(self):
        ActionChains(self.driver.browser).context_click(self.native).perform()

    def select_option(self):
        if not self.selected:
            self.native.click()

    def send_keys(self, *args):
        self.native.send_keys(*args)

    def set(self, value):
        tag_name = self.tag_name
        type_attr = self["type"]

        if tag_name == "input" and type_attr == "radio":
            self.click()
        elif tag_name == "input" and type_attr == "checkbox":
            current = self.native.get_attribute("checked") == "true"

            if current ^ value:
                self.click()
        elif tag_name == "textarea" or tag_name == "input":
            if self.readonly:
                raise ReadOnlyElementError()

            # Clear field by JavaScript assignment of the value property.
            self.driver.browser.execute_script("arguments[0].value = ''", self.native)
            self.native.send_keys(value)
        elif self["isContentEditable"]:
            self.native.click()
            script = """
                var range = document.createRange();
                var sel = window.getSelection();
                range.selectNodeContents(arguments[0]);
                sel.removeAllRanges();
                sel.addRange(range);
            """
            self.driver.browser.execute_script(script, self.native)
            self.native.send_keys(value)

    def unselect_option(self):
        if not self._select_node.multiple:
            raise UnselectNotAllowed("Cannot unselect option from single select box.")

        if self.selected:
            self.native.click()

    def __eq__(self, other):
        return self.native == other.native

    def __hash__(self):
        return hash(self.native)

    @property
    def checked(self):
        return self.selected

    @property
    def selected(self):
        return self.native.is_selected()

    @property
    def disabled(self):
        return not self.native.is_enabled()

    @property
    def readonly(self):
        val = self["readonly"]
        return bool(val) and val != "false"

    @property
    def _select_node(self):
        return self._find_xpath("./ancestor::select")[0]



class Window(object):
    """
    The Window class represents a browser window.

    You can get an instance of the class by calling either of:

    * :attr:`windows`
    * :attr:`current_window`
    * :meth:`window_opened_by`
    * :meth:`switch_to_window`

    Note that some drivers (e.g. Selenium) support getting size of/resizing/closing only current
    window. So if you invoke such method for:

    * window that is current, Capybara will make 2 Selenium method invocations (get handle of
      current window + get size/resize/close).
    * window that is not current, Capybara will make 4 Selenium method invocations (get handle of
      current window + switch to given handle + get size/resize/close + switch to original handle)

    Args:
        session (Session): Session that this window belongs to.
        handle (object): An object that uniquely identifies the window within the session.
    """

    def __init__(self, session, handle):
        self.session = session
        self.driver = session.driver
        self.handle = handle

    def __repr__(self):
        return "<capybara.window.Window handle={}>".format(repr(self.handle))

    def __eq__(self, other):
        return (
            type(self) == type(other) and
            self.session == other.session and
            self.handle == other.handle)

    def __hash__(self):
        return hash(self.session) ^ hash(self.handle)

    @property
    def closed(self):
        """ bool: Whether the window is closed. """
        return not self.exists

    @property
    def current(self):
        """ bool: Whether this window is the window in which commands are being executed. """
        try:
            return self.driver.current_window_handle == self.handle
        except self.driver.no_such_window_error:
            return False

    @property
    def exists(self):
        """ bool: Whether the window is not closed. """
        return self.handle in self.driver.window_handles

    @property
    def size(self):
        """ List[int, int]: The window size. """
        return self.driver.window_size(self.handle)

    def close(self):
        """
        Close window.

        If this method was called for the window that is current, then after calling this method
        future invocations of other Capybara methods should raise
        ``session.driver.no_such_window_error`` until another window is switched to.

        If this method was called for a window that is not current, then after calling this method
        the current window should remain the same as it was before calling this method.
        """

        self.driver.close_window(self.handle)

    def maximize(self):
        """
        Maximize the window.

        If this method was called for a window that is not current, then after calling this method
        the current window should remain the same as it was before calling this method.
        """

        self.driver.maximize_window(self.handle)

    def resize_to(self, width, height):
        """
        Resizes the window to the given dimensions.

        If this method was called for a window that is not current, then after calling this method
        the current window should remain the same as it was before calling this method.

        Args:
            width (int): The new window width in pixels.
            height (int): The new window height in pixels.
        """

        self.driver.resize_window_to(self.handle, width, height)



import pytest

from capybara.exceptions import ElementNotFound
from capybara.tests.helpers import extract_results


class TestScope:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_scope")

    def test_asserts_content_in_the_given_scope(self, session):
        with session.scope("css", "#for_foo"):
            assert not session.has_text("First Name")
        assert session.has_text("First Name")

    def test_accepts_additional_options(self, session):
        with session.scope("css", "#for_bar li", text="With Simple HTML"):
            session.click_link("Go")
        assert session.has_text("Bar")

    def test_clicks_links_in_the_given_xpath_scope(self, session):
        with session.scope("xpath", "//div[@id='for_bar']//li[contains(.,'With Simple HTML')]"):
            session.click_link("Go")
        assert session.has_text("Bar")

    def test_clicks_links_in_the_given_node(self, session):
        node_of_interest = session.find("css", "#for_bar li", text="With Simple HTML")
        with session.scope(node_of_interest):
            session.click_link("Go")
        assert session.has_text("Bar")

    def test_respects_the_inner_scope_of_nested_scopes(self, session):
        with session.scope("//div[@id='for_bar']"):
            with session.scope(".//li[contains(.,'Bar')]"):
                session.click_link("Go")
        assert session.has_text("Another World")

    def test_respects_the_outer_scope_of_nested_scopes(self, session):
        with session.scope("//div[@id='another_foo']"):
            with session.scope(".//li[contains(.,'With Simple HTML')]"):
                session.click_link("Go")
        assert session.has_text("Hello world")

    def test_raises_an_error_if_the_scope_is_not_found_on_the_page(self, session):
        with pytest.raises(ElementNotFound):
            with session.scope("//div[@id='doesnotexist']"):
                pass

    def test_restores_the_scope_when_an_error_is_raised(self, session):
        assert session.has_xpath(".//div[@id='another_foo']")
        with session.scope("//div[@id='for_bar']"):
            assert not session.has_xpath(".//div[@id='another_foo']")
            with pytest.raises(ElementNotFound):
                with session.scope(".//div[@id='doesnotexist']"):
                    pass
            assert not session.has_xpath(".//div[@id='another_foo']")
        assert session.has_xpath(".//div[@id='another_foo']")

    def test_fills_in_a_field_and_clicks_a_button(self, session):
        with session.scope("//li[contains(.,'Bar')]"):
            session.click_button("Go")
        assert extract_results(session)["form[first_name]"] == "Peter"

        session.visit("/with_scope")
        with session.scope("//li[contains(.,'Bar')]"):
            session.fill_in("First Name", value="Dagobert")
            session.click_button("Go")
        assert extract_results(session)["form[first_name]"] == "Dagobert"



from functools import wraps
from time import sleep, time

import capybara
from capybara.exceptions import ElementNotFound, FrozenInTime
from capybara.node.actions import ActionsMixin
from capybara.node.finders import FindersMixin
from capybara.node.matchers import MatchersMixin


class Base(FindersMixin, ActionsMixin, MatchersMixin, object):
    """
    A :class:`Base` represents either an element on a page through the subclass :class:`Element` or
    a document through :class:`Document`.

    Both types of Node share the same methods, used for interacting with the elements on the page.
    These methods are divided into three categories: finders, actions, and matchers. These are found
    in the classes :class:`FindersMixin`, :class:`ActionsMixin`, and :class:`MatchersMixin`
    respectively.

    A :class:`Session` exposes all methods from :class:`Document` directly::

        session = Session("selenium", my_app)
        session.visit("/")
        session.fill_in("Foo", value="Bar")  # from capybara.node.actions.ActionsMixin
        bar = session.find("#bar")           # from capybara.node.finders.FindersMixin
        bar.select("Baz", field="Quox")      # from capybara.node.actions.ActionsMixin
        session.has_css("#foobar")           # from capybara.node.matchers.MatchersMixin

    Args:
        session (Session): The session from which this node originated.
        base (driver.Node): The underlying driver node.
    """

    def __init__(self, session, base):
        self.session = session
        self.base = base
        self.allow_reload = False

    def reload(self):
        """
        Reloads the underlying driver node.

        Returns:
            node.Base: This node.
        """

        return self

    def __eq__(self, other):
        return id(self) == id(other) or (hasattr(other, "base") and self.base == other.base)

    def __hash__(self):
        return hash(self.base)

    def __getitem__(self, name):
        """
        Retrieve the given attribute.

        Args:
            name (str): The attribute to retrieve.

        Returns:
            str: The value of the attribute.
        """

        raise NotImplementedError()

    @property
    def text(self):
        """ str: The text of the node. """
        raise NotImplementedError()

    @property
    def all_text(self):
        """ str: All of the text of the node. """
        raise NotImplementedError()

    @property
    def visible_text(self):
        """ str: Only the visible text of the node. """
        raise NotImplementedError()

    def synchronize(self, func=None, wait=None, errors=()):
        """
        This method is Capybara's primary defense against asynchronicity problems. It works by
        attempting to run a given decorated function until it succeeds. The exact behavior of this
        method depends on a number of factors. Basically there are certain exceptions which, when
        raised from the decorated function, instead of bubbling up, are caught, and the function is
        re-run.

        Certain drivers have no support for asynchronous processes. These drivers run the function,
        and any error raised bubbles up immediately. This allows faster turn around in the case
        where an expectation fails.

        Only exceptions that are :exc:`ElementNotFound` or any subclass thereof cause the block to
        be rerun. Drivers may specify additional exceptions which also cause reruns. This usually
        occurs when a node is manipulated which no longer exists on the page. For example, the
        Selenium driver specifies ``selenium.common.exceptions.StateElementReferenceException``.

        As long as any of these exceptions are thrown, the function is re-run, until a certain
        amount of time passes. The amount of time defaults to :data:`capybara.default_max_wait_time`
        and can be overridden through the ``wait`` argument. This time is compared with the system
        time to see how much time has passed. If the return value of ``time.time()`` is stubbed
        out, Capybara will raise :exc:`FrozenInTime`.

        Args:
            func (Callable, optional): The function to decorate.
            wait (int, optional): Number of seconds to retry this function.
            errors (Tuple[Type[Exception]], optional): Exception types that cause the function to be
                rerun. Defaults to ``driver.invalid_element_errors`` + :exc:`ElementNotFound`.

        Returns:
            Callable: The decorated function, or a decorator function.

        Raises:
            FrozenInTime: If the return value of ``time.time()`` appears stuck.
        """

        def decorator(func):
            @wraps(func)
            def outer(*args, **kwargs):
                caught_errors = (
                    errors or
                    self.session.driver.invalid_element_errors + (ElementNotFound,))
                seconds = wait if wait is not None else capybara.default_max_wait_time

                def inner():
                    return func(*args, **kwargs)

                if self.session.synchronized:
                    return inner()
                else:
                    start_time = time()
                    self.session.synchronized = True
                    try:
                        while True:
                            try:
                                return inner()
                            except Exception as e:
                                if not isinstance(e, caught_errors):
                                    raise
                                if time() - start_time >= seconds:
                                    raise

                                sleep(0.05)

                                if time() == start_time:
                                    raise FrozenInTime(
                                        "time appears to be frozen, Capybara does not work with "
                                        "libraries which freeze time, consider using time "
                                        "traveling instead")
                                if capybara.automatic_reload:
                                    self.reload()
                    finally:
                        self.session.synchronized = False

            return outer

        if func:
            return decorator(func)
        else:
            return decorator

    def _find_css(self, css):
        return self.base._find_css(css)

    def _find_xpath(self, xpath):
        return self.base._find_xpath(xpath)


def synchronize(func):
    """ Decorator for :meth:`synchronize`. """

    @wraps(func)
    def outer(self, *args, **kwargs):
        @self.synchronize
        def inner(self, *args, **kwargs):
            return func(self, *args, **kwargs)

        return inner(self, *args, **kwargs)

    return outer



from contextlib import contextmanager
from datetime import datetime
from functools import wraps
import os
import random

import capybara
from capybara.compat import ParseResult, urlparse
from capybara.driver.node import Node
from capybara.exceptions import ScopeError, WindowError
from capybara.node.base import Base
from capybara.node.document import Document
from capybara.node.element import Element
from capybara.server import Server
from capybara.session_matchers import SessionMatchersMixin
from capybara.utils import cached_property, encode_string
from capybara.window import Window


_DOCUMENT_METHODS = ["assert_no_title", "assert_title", "has_no_title", "has_title"]
_DOCUMENT_PROPERTIES = ["title"]
_NODE_METHODS = [
    "assert_no_selector", "assert_no_text", "assert_selector", "assert_text", "attach_file",
    "check", "choose", "click_button", "click_link", "click_link_or_button", "click_on", "fill_in",
    "find", "find_all", "find_button", "find_by_id", "find_field", "find_first", "find_link",
    "has_button", "has_checked_field", "has_content", "has_css", "has_field", "has_link",
    "has_no_button", "has_no_checked_field", "has_no_css", "has_no_field", "has_no_link",
    "has_no_select", "has_no_selector", "has_no_table", "has_no_text", "has_no_unchecked_field",
    "has_no_xpath", "has_select", "has_selector", "has_table", "has_text", "has_unchecked_field",
    "has_xpath", "select", "uncheck", "unselect"]
_NODE_PROPERTIES = ["text"]
_SESSION_METHODS = [
    "accept_alert", "accept_confirm", "accept_prompt", "assert_current_path",
    "assert_no_current_path", "dismiss_confirm", "dismiss_prompt", "evaluate_script",
    "execute_script", "fieldset", "frame", "go_back", "go_forward", "has_current_path",
    "has_no_current_path", "open_new_window", "reset", "save_page", "save_screenshot", "scope",
    "switch_to_window", "table", "visit", "window", "window_opened_by"]
_SESSION_PROPERTIES = ["current_host", "current_path", "current_url", "current_window", "windows"]

DSL_METHODS = _DOCUMENT_METHODS + _NODE_METHODS + _SESSION_METHODS


class Session(SessionMatchersMixin, object):
    """
    The Session class represents a single user's interaction with the system. The Session can use
    any of the underlying drivers. A session can be initialized manually like this::

        session = Session("selenium", MyWSGIApp)

    The application given as the second argument is optional. When running Capybara against an
    external page, you might want to leave it out::

        session = Session("selenium")
        session.visit("http://www.google.com")

    Session provides a number of methods and properties for controlling the navigation of the page,
    such as :meth:`visit`, :attr:`current_path`, and so on. It also delegates a number of methods to
    a :class:`Document`, representing the current HTML document. This allows interaction::

        session.fill_in("q", value="Capybara")
        session.click_button("Search")
        assert session.has_text("Capybara")

    When using :mod:`capybara.dsl`, the Session is initialized automatically for you.

    Args:
        mode (str): The name of the driver to use.
        app (object): The WSGI-compliant app with which to interact.
    """

    def __init__(self, mode, app):
        self.mode = mode
        self.app = app
        self.server = Server(app).boot() if app and self.driver.needs_server else None
        self.synchronized = False
        self._scopes = [None]

    @cached_property
    def driver(self):
        """ driver.Base: The driver for the current session. """
        return capybara.drivers[self.mode](self.app)

    @cached_property
    def document(self):
        """ Document: The document for the current page. """
        return Document(self, self.driver)

    @property
    def current_scope(self):
        """ node.Base: The current node relative to which all interaction will be scoped. """
        return self._scopes[-1] or self.document

    @property
    def html(self):
        """ str: A snapshot of the DOM of the current document, as it looks right now. """
        return self.driver.html

    body = html
    """ Alias for :attr:`html`. """

    source = html
    """ Alias for :attr:`html`. """

    @property
    def current_path(self):
        """ str: Path of the current page, without any domain information. """

        if not self.current_url:
            return

        path = urlparse(self.current_url).path
        return path if path else None

    @property
    def current_host(self):
        """ str: Host of the current page. """

        if not self.current_url:
            return

        result = urlparse(self.current_url)
        scheme, netloc = result.scheme, result.netloc
        host = netloc.split(":")[0] if netloc else None
        return "{0}://{1}".format(scheme, host) if host else None

    @property
    def current_url(self):
        """ str: Fully qualified URL of the current page. """
        return self.driver.current_url

    def visit(self, visit_uri):
        """
        Navigate to the given URL. The URL can either be a relative URL or an absolute URL. The
        behavior of either depends on the driver. ::

            session.visit("/foo")
            session.visit("http://google.com")

        For drivers which can run against an external application, such as the Selenium driver,
        giving an absolute URL will navigate to that page. This allows testing applications running
        on remote servers. For these drivers, setting :data:`capybara.app_host` will make the
        remote server the default. For example::

            capybara.app_host = "http://google.com"
            session.visit("/")  # visits the Google homepage

        Args:
            visit_uri (str): The URL to navigate to.
        """

        visit_uri = urlparse(visit_uri)

        if capybara.app_host:
            uri_base = urlparse(capybara.app_host)
        elif self.server:
            uri_base = urlparse("http://{}:{}".format(self.server.host, self.server.port))
        else:
            uri_base = None

        visit_uri = ParseResult(
            scheme=visit_uri.scheme or (uri_base.scheme if uri_base else ""),
            netloc=visit_uri.netloc or (uri_base.netloc if uri_base else ""),
            path=visit_uri.path,
            params=visit_uri.params,
            query=visit_uri.query,
            fragment=visit_uri.fragment)

        self.driver.visit(visit_uri.geturl())

    def go_back(self):
        """ Move back a single entry in the browser's history. """
        self.driver.go_back()

    def go_forward(self):
        """ Move forward a single entry in the browser's history. """
        self.driver.go_forward()

    @contextmanager
    def scope(self, *args, **kwargs):
        """
        Executes the wrapped code within the context of a node. ``scope`` takes the same options
        as :meth:`find`. For the duration of the context, any command to Capybara will be handled
        as though it were scoped to the given element. ::

            with scope("xpath", "//div[@id='delivery-address']"):
                fill_in("Street", value="12 Main Street")

        Just as with :meth:`find`, if multiple elements match the selector given to ``scope``, an
        error will be raised, and just as with :meth:`find`, this behavior can be controlled
        through the ``match`` and ``exact`` options.

        It is possible to omit the first argument, in that case, the selector is assumed to be of
        the type set in :data:`capybara.default_selector`. ::

            with scope("div#delivery-address"):
                fill_in("Street", value="12 Main Street")

        Note that a lot of uses of ``scope`` can be replaced more succinctly with chaining::

            find("div#delivery-address").fill_in("Street", value="12 Main Street")

        Args:
            *args: Variable length argument list for the call to :meth:`find`.
            **kwargs: Arbitrary keywords arguments for the call to :meth:`find`.
        """

        new_scope = args[0] if isinstance(args[0], Base) else self.find(*args, **kwargs)
        self._scopes.append(new_scope)
        try:
            yield
        finally:
            self._scopes.pop()

    @contextmanager
    def fieldset(self, locator):
        """
        Execute the wrapped code within a specific fieldset given the id or legend of that
        fieldset.

        Args:
            locator (str): The id or legend of the fieldset.
        """

        with self.scope("fieldset", locator):
            yield

    @contextmanager
    def table(self, locator):
        """
        Execute the wrapped code within a specific table given the id or caption of that table.

        Args:
            locator (str): The id or caption of the table.
        """

        with self.scope("table", locator):
            yield

    @contextmanager
    def frame(self, locator):
        """
        Execute the wrapped code within the given iframe using the given frame or frame name/id.
        May not be supported by all drivers.

        Args:
            locator (str | Element): The name/id of the frame or the frame's element.
        """

        self._scopes.append(None)
        try:
            new_frame = locator if isinstance(locator, Element) else self.find("frame", locator)
            self.driver.switch_to_frame(new_frame)
            try:
                yield
            finally:
                self.driver.switch_to_frame("parent")
        finally:
            self._scopes.pop()

    @property
    def current_window(self):
        """ Window: The current window. """
        return Window(self, self.driver.current_window_handle)

    @property
    def windows(self):
        """
        Get all opened windows. The order of the windows in the returned list is not defined. The
        driver may sort windows by their creation time but it's not required.

        Returns:
            List[Window]: A list of all windows.
        """

        return [Window(self, window_handle) for window_handle in self.driver.window_handles]

    def open_new_window(self):
        """
        Open new window. The current window doesn't change as a result of this call. It should be
        switched explicitly.

        Returns:
            Window: The window that has been opened.
        """

        return self.window_opened_by(lambda: self.driver.open_new_window())

    def switch_to_window(self, window, wait=None):
        """
        If ``window`` is a lambda, it switches to the first window for which ``window`` returns a
        value other than False or None. If a window that matches can't be found, the window will be
        switched back and :exc:`WindowError` will be raised.

        Args:
            window (Window | lambda): The window that should be switched to, or a filtering lambda.
            wait (int | float, optional): The number of seconds to wait to find the window.

        Returns:
            Window: The new current window.

        Raises:
            ScopeError: If this method is invoked inside :meth:`scope, :meth:`frame`, or
                :meth:`window`.
            WindowError: If no window matches the given lambda.
        """

        if len(self._scopes) > 1:
            raise ScopeError(
                "`switch_to_window` is not supposed to be invoked from "
                "within `scope`s, `frame`s, or other `window`s.")

        if isinstance(window, Window):
            self.driver.switch_to_window(window.handle)
            return window
        else:
            @self.document.synchronize(errors=(WindowError,), wait=wait)
            def switch_and_get_matching_window():
                original_window_handle = self.driver.current_window_handle
                try:
                    for handle in self.driver.window_handles:
                        self.driver.switch_to_window(handle)
                        result = window()
                        if result:
                            return Window(self, handle)
                except Exception:
                    self.driver.switch_to_window(original_window_handle)
                    raise

                self.driver.switch_to_window(original_window_handle)
                raise WindowError("Could not find a window matching lambda")

            return switch_and_get_matching_window()

    @contextmanager
    def window(self, window):
        """
        This method does the following:

        1. Switches to the given window (it can be located by window instance/lambda/string).
        2. Executes the given block (within window located at previous step).
        3. Switches back (this step will be invoked even if exception happens at second step).

        Args:
            window (Window | lambda): The desired :class:`Window`, or a lambda that will be run in
                the context of each open window and returns ``True`` for the desired window.
        """

        original = self.current_window
        if window != original:
            self.switch_to_window(window)
        self._scopes.append(None)
        try:
            yield
        finally:
            self._scopes.pop()
            if original != window:
                self.switch_to_window(original)

    def window_opened_by(self, trigger_func, wait=None):
        """
        Get the window that has been opened by the passed lambda. It will wait for it to be opened
        (in the same way as other Capybara methods wait). It's better to use this method than
        ``windows[-1]`` `as order of windows isn't defined in some drivers`__.

        __ https://dvcs.w3.org/hg/webdriver/raw-file/default/webdriver-spec.html#h_note_10

        Args:
            trigger_func (func): The function that should trigger the opening of a new window.
            wait (int | float, optional): Maximum wait time. Defaults to
                :data:`capybara.default_max_wait_time`.

        Returns:
            Window: The window that has been opened within the lambda.

        Raises:
            WindowError: If lambda passed to window hasn't opened window or opened more than one
                window.
        """

        old_handles = set(self.driver.window_handles)
        trigger_func()

        @self.document.synchronize(wait=wait, errors=(WindowError,))
        def get_new_window():
            opened_handles = set(self.driver.window_handles) - old_handles
            if len(opened_handles) != 1:
                raise WindowError("lambda passed to `window_opened_by` "
                                  "opened {0} windows instead of 1".format(len(opened_handles)))
            return Window(self, list(opened_handles)[0])

        return get_new_window()

    def execute_script(self, script, *args):
        """
        Execute the given script, not returning a result. This is useful for scripts that return
        complex objects, such as jQuery statements. ``execute_script`` should be used over
        :meth:`evaluate_script` whenever possible.

        Args:
            script (str): A string of JavaScript to execute.
            *args: Variable length argument list to pass to the executed JavaScript string.
        """

        args = [arg.base if isinstance(arg, Base) else arg for arg in args]
        self.driver.execute_script(script, *args)

    def evaluate_script(self, script, *args):
        """
        Evaluate the given JavaScript and return the result. Be careful when using this with
        scripts that return complex objects, such as jQuery statements. :meth:`execute_script`
        might be a better alternative.

        Args:
            script (str): A string of JavaScript to evaluate.
            *args: Variable length argument list to pass to the executed JavaScript string.

        Returns:
            object: The result of the evaluated JavaScript (may be driver specific).
        """

        args = [arg.base if isinstance(arg, Base) else arg for arg in args]
        result = self.driver.evaluate_script(script, *args)
        return self._wrap_element_script_result(result)

    @contextmanager
    def accept_alert(self, text=None, wait=None):
        """
        Execute the wrapped code, accepting an alert.

        Args:
            text (str, optional): Text to match against the text in the modal.
            wait (int | float, optional): Maximum time to wait for the modal to appear after
                executing the wrapped code.

        Raises:
            ModalNotFound: If a modal dialog hasn't been found.
        """

        wait = wait or capybara.default_max_wait_time
        with self.driver.accept_modal("alert", text=text, wait=wait):
            yield

    @contextmanager
    def accept_confirm(self, text=None, wait=None):
        """
        Execute the wrapped code, accepting a confirm.

        Args:
            text (str, optional): Text to match against the text in the modal.
            wait (int | float, optional): Maximum time to wait for the modal to appear after
                executing the wrapped code.

        Raises:
            ModalNotFound: If a modal dialog hasn't been found.
        """

        with self.driver.accept_modal("confirm", text=text, wait=wait):
            yield

    @contextmanager
    def dismiss_confirm(self, text=None, wait=None):
        """
        Execute the wrapped code, dismissing a confirm.

        Args:
            text (str, optional): Text to match against the text in the modal.
            wait (int | float, optional): Maximum time to wait for the modal to appear after
                executing the wrapped code.

        Raises:
            ModalNotFound: If a modal dialog hasn't been found.
        """

        with self.driver.dismiss_modal("confirm", text=text, wait=wait):
            yield

    @contextmanager
    def accept_prompt(self, text=None, response=None, wait=None):
        """
        Execute the wrapped code, accepting a prompt, optionally responding to the prompt.

        Args:
            text (str, optional): Text to match against the text in the modal.
            response (str, optional): Response to provide to the prompt.
            wait (int | float, optional): Maximum time to wait for the modal to appear after
                executing the wrapped code.

        Raises:
            ModalNotFound: If a modal dialog hasn't been found.
        """

        with self.driver.accept_modal("prompt", text=text, response=response, wait=wait):
            yield

    @contextmanager
    def dismiss_prompt(self, text=None, wait=None):
        """
        Execute the wrapped code, dismissing a prompt.

        Args:
            text (str, optional): Text to match against the text in the modal.
            wait (int | float, optional): Maximum time to wait for the modal to appear after
                executing the wrapped code.

        Raises:
            ModalNotFound: If a modal dialog hasn't been found.
        """

        with self.driver.dismiss_modal("prompt", text=text, wait=wait):
            yield

    def save_page(self, path=None):
        """
        Save a snapshot of the page.

        If invoked without arguments, it will save a file to :data:`capybara.save_path` and the
        file will be given a randomly generated filename. If invoked with a relative path, the path
        will be relative to :data:`capybara.save_path`.

        Args:
            path (str, optional): The path to where it should be saved.

        Returns:
            str: The path to which the file was saved.
        """

        path = _prepare_path(path, "html")

        with open(path, "wb") as f:
            f.write(encode_string(self.body))

        return path

    def save_screenshot(self, path=None, **kwargs):
        """
        Save a screenshot of the page.

        If invoked without arguments, it will save a file to :data:`capybara.save_path` and the
        file will be given a randomly generated filename. If invoked with a relative path, the path
        will be relative to :data:`capybara.save_path`.

        Args:
            path (str, optional): The path to where it should be saved.
            **kwargs: Arbitrary keywords arguments for the driver.

        Returns:
            str: The path to which the file was saved.
        """

        path = _prepare_path(path, "png")
        self.driver.save_screenshot(path, **kwargs)
        return path

    def reset(self):
        """
        Reset the session (i.e., remove cookies and navigate to a blank page).

        This method does not:
        * accept modal dialogs if they are present (the Selenium driver does, but others may not),
        * clear the browser cache/HTML 5 local storage/IndexedDB/Web SQL database/etc., or
        * modify the state of the driver/underlying browser in any other way

        as doing so would result in performance downsides and it's not needed to do everything
        from the list above for most apps.

        If you want to do anything from the list above on a general basis you can write a test
        teardown method.
        """

        self.driver.reset()

    cleanup = reset
    """ Alias for :meth:`reset`. """

    reset_session = reset
    """ Alias for :meth:`reset`. """

    def _wrap_element_script_result(self, arg):
        if isinstance(arg, list):
            return [self._wrap_element_script_result(e) for e in arg]
        elif isinstance(arg, Node):
            return Element(self, arg, None, None)
        else:
            return arg


def _define_document_method(method_name):
    @wraps(getattr(Document, method_name))
    def func(self, *args, **kwargs):
        return getattr(self.document, method_name)(*args, **kwargs)
    setattr(Session, method_name, func)


def _define_document_property(property_name):
    def fget(self):
        return getattr(self.document, property_name)
    fdoc = getattr(Document, property_name).__doc__
    setattr(Session, property_name, property(fget, None, None, fdoc))


def _define_node_method(method_name):
    @wraps(getattr(Base, method_name))
    def func(self, *args, **kwargs):
        return getattr(self.current_scope, method_name)(*args, **kwargs)
    setattr(Session, method_name, func)


def _define_node_property(property_name):
    def fget(self):
        return getattr(self.current_scope, property_name)
    fdoc = getattr(Base, property_name).__doc__
    setattr(Session, property_name, property(fget, None, None, fdoc))


for method_name in _DOCUMENT_METHODS:
    _define_document_method(method_name)

for property_name in _DOCUMENT_PROPERTIES:
    _define_document_property(property_name)

for method_name in _NODE_METHODS:
    _define_node_method(method_name)

for property_name in _NODE_PROPERTIES:
    _define_node_property(property_name)


def _prepare_path(path, extension):
    save_path = capybara.save_path or os.getcwd()
    path = os.path.normpath(os.path.join(save_path, path or _default_fn(extension)))

    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))

    return path


def _default_fn(extension):
    timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
    return "capybara-{timestamp}{random}.{extension}".format(
        timestamp=timestamp,
        random=random.randint(0, 10**10),
        extension=extension)



import pytest


@pytest.mark.requires("frames")
class TestFrame:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/within_frames")

    def test_finds_the_div_in_frame_one(self, session):
        with session.frame("frameOne"):
            assert session.find("//*[@id='divInFrameOne']").text == "This is the text of divInFrameOne"

    def test_finds_the_div_in_frame_two(self, session):
        with session.frame("frameTwo"):
            assert session.find("//*[@id='divInFrameTwo']").text == "This is the text of divInFrameTwo"

    def test_finds_the_text_div_in_the_main_window_after_finding_text_in_frame_one(self, session):
        with session.frame("frameOne"):
            assert session.find("//*[@id='divInFrameOne']").text == "This is the text of divInFrameOne"
        assert session.find("//*[@id='divInMainWindow']").text == "This is the text for divInMainWindow"

    def test_finds_the_text_div_in_the_main_window_after_finding_text_in_frame_two(self, session):
        with session.frame("frameTwo"):
            assert session.find("//*[@id='divInFrameTwo']").text == "This is the text of divInFrameTwo"
        assert session.find("//*[@id='divInMainWindow']").text == "This is the text for divInMainWindow"

    def test_finds_the_div_given_an_element(self, session):
        element = session.find("css", "#frameOne")
        with session.frame(element):
            assert session.find("//*[@id='divInFrameOne']").text == "This is the text of divInFrameOne"

    def test_finds_multiple_nested_frames(self, session):
        with session.frame("parentFrame"):
            with session.frame("childFrame"):
                with session.frame("grandchildFrame1"):
                    pass
                with session.frame("grandchildFrame2"):
                    pass

    def test_handles_a_frame_closing(self, session):
        with session.frame("parentFrame"):
            with session.frame("childFrame"):
                session.click_link("Close Window")
            assert session.has_selector("css", "body#parentBody")
            assert not session.has_selector("css", "#childFrame")



import os
import pytest
import re
import shutil

import capybara
from capybara.tests.compat import patch


@pytest.mark.requires("screenshot")
class TestSaveScreenshot:
    @pytest.fixture
    def alternative_path(self):
        return os.path.join(os.getcwd(), "save_screenshot_tmp")

    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/foo")

    @pytest.fixture(autouse=True)
    def teardown_settings_and_files(self, alternative_path):
        yield

        capybara.save_path = None

        if os.path.isdir(alternative_path):
            shutil.rmtree(alternative_path)

    def test_generates_sensible_filename(self, session):
        with patch.object(session.driver, "save_screenshot") as save_screenshot:
            session.save_screenshot()

            regex = re.compile(os.path.abspath("capybara-\d+\.png"))

            assert len(save_screenshot.call_args_list) == 1
            assert regex.search(save_screenshot.call_args[0][0])

    def test_allows_one_to_specify_another_path(self, session, alternative_path):
        with patch.object(session.driver, "save_screenshot") as save_screenshot:
            capybara.save_path = alternative_path

            custom_path = "screenshots/1.png"
            session.save_screenshot(custom_path)

            regex = re.compile(re.escape(custom_path) + "$")

            assert len(save_screenshot.call_args_list) == 1
            assert regex.search(save_screenshot.call_args[0][0])

    def test_saves_file_in_custom_save_path(self, session, alternative_path):
        with patch.object(session.driver, "save_screenshot") as save_screenshot:
            capybara.save_path = alternative_path

            session.save_screenshot()

            regex = re.compile(os.path.normpath(os.path.join(alternative_path, "capybara-\d+\.png")))

            assert len(save_screenshot.call_args_list) == 1
            assert regex.search(save_screenshot.call_args[0][0])

    def test_saves_file_relative_to_custom_save_path(self, session, alternative_path):
        with patch.object(session.driver, "save_screenshot") as save_screenshot:
            capybara.save_path = alternative_path

            custom_path = "screenshots/1.png"
            session.save_screenshot(custom_path)

            regex = re.compile(os.path.normpath(os.path.join(alternative_path, re.escape(custom_path))))

            assert len(save_screenshot.call_args_list) == 1
            assert regex.search(save_screenshot.call_args[0][0])



import pytest

from capybara.exceptions import ElementNotFound, UnselectNotAllowed
from capybara.tests.helpers import extract_results


class TestUnselect:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")

    def test_raises_an_error_with_single_select(self, session):
        with pytest.raises(UnselectNotAllowed):
            session.unselect("English", field="form_locale")

    def test_raises_an_error_with_a_locator_that_does_not_exist(self, session):
        with pytest.raises(ElementNotFound):
            session.unselect("foo", field="does not exist")

    def test_raises_an_error_with_an_option_that_does_not_exist(self, session):
        with pytest.raises(ElementNotFound):
            session.unselect("Does not Exist", field="form_underwear")

    def test_approximately_matches_select_box(self, session):
        session.unselect("Boxerbriefs", field="Under")
        session.click_button("awesome")
        underwear = extract_results(session).getlist("form[underwear][]")
        assert "Boxerbriefs" not in underwear

    def test_approximately_matches_option(self, session):
        session.unselect("Boxerbr", field="Underwear")
        session.click_button("awesome")
        underwear = extract_results(session).getlist("form[underwear][]")
        assert "Boxerbriefs" not in underwear

    def test_approximately_matches_when_field_not_given(self, session):
        session.unselect("Boxerbr")
        session.click_button("awesome")
        underwear = extract_results(session).getlist("form[underwear][]")
        assert "Boxerbriefs" not in underwear

    def test_does_not_approximately_match_select_box(self, session):
        with pytest.raises(ElementNotFound):
            session.unselect("Boxerbriefs", field="Under", exact=True)

    def test_does_not_approximately_match_option(self, session):
        with pytest.raises(ElementNotFound):
            session.unselect("Boxerbr", field="Underwear", exact=True)

    def test_does_not_approximately_match_when_field_not_given(self, session):
        with pytest.raises(ElementNotFound):
            session.unselect("Boxerbr", exact=True)


class TestMultipleUnselect:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")

    def test_unselects_an_option_from_a_select_box_by_id(self, session):
        session.unselect("Commando", field="form_underwear")
        session.click_button("awesome")
        underwear = extract_results(session).getlist("form[underwear][]")
        assert "Briefs" in underwear
        assert "Boxerbriefs" in underwear
        assert "Command" not in underwear

    def test_unselects_an_option_without_a_select_box(self, session):
        session.unselect("Commando")
        session.click_button("awesome")
        underwear = extract_results(session).getlist("form[underwear][]")
        assert "Briefs" in underwear
        assert "Boxerbriefs" in underwear
        assert "Command" not in underwear

    def test_unselects_an_option_from_a_select_box_by_label(self, session):
        session.unselect("Commando", field="Underwear")
        session.click_button("awesome")
        underwear = extract_results(session).getlist("form[underwear][]")
        assert "Briefs" in underwear
        assert "Boxerbriefs" in underwear
        assert "Command" not in underwear

    def test_escapes_quotes(self, session):
        session.unselect("Frenchman's Pantalons", field="Underwear")
        session.click_button("awesome")
        underwear = extract_results(session).getlist("form[underwear][]")
        assert "Frenchman's Pantalons" not in underwear



import pytest


class HasFieldTestCase:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")


class TestHasField(HasFieldTestCase):
    def test_is_true_if_the_field_is_on_the_page(self, session):
        assert session.has_field("Dog")
        assert session.has_field("form_description")
        assert session.has_field("Region")

    def test_is_false_if_the_field_is_not_on_the_page(self, session):
        assert not session.has_field("Monkey")


class TestHasFieldDisabled(HasFieldTestCase):
    def test_does_not_find_disabled_fields_when_false(self, session):
        assert not session.has_field("Disabled Checkbox", disabled=False)

    def test_finds_enabled_fields_when_false(self, session):
        assert session.has_field("Dog", disabled=False)

    def test_finds_disabled_fields_when_true(self, session):
        assert session.has_field("Disabled Checkbox", disabled=True)

    def test_does_not_find_enabled_fields_when_true(self, session):
        assert not session.has_field("Dog", disabled=True)

    def test_finds_disabled_fields_when_all(self, session):
        assert session.has_field("Disabled Checkbox", disabled="all")

    def test_finds_enabled_fields_when_all(self, session):
        assert session.has_field("Dog", disabled="all")


class TestHasNoField:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")

    def test_is_false_if_the_field_is_on_the_page(self, session):
        assert not session.has_no_field("Dog")
        assert not session.has_no_field("form_description")
        assert not session.has_no_field("Region")

    def test_is_true_if_the_field_is_not_on_the_page(self, session):
        assert session.has_no_field("Monkey")



import pytest

from capybara.node.document import Document


class TestCurrentScope:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_scope")

    def test_returns_the_document_when_outside_of_a_scope(self, session):
        assert isinstance(session.current_scope, Document)

    def test_returns_the_element_in_scope(self, session):
        with session.scope("//*[@id='simple_first_name']"):
            assert session.current_scope["name"] == "first_name"

    def test_returns_the_element_in_nested_scope(self, session):
        with session.scope("//div[@id='for_bar']"):
            with session.scope(".//input[@value='Peter']"):
                assert session.current_scope["name"] == "form[first_name]"



import pytest
import re


class TestHasCSS:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")

    def test_is_true_if_the_given_selector_is_on_the_page(self, session):
        assert session.has_css("p")
        assert session.has_css("p a#foo")

    def test_is_false_if_the_given_selector_is_not_on_the_page(self, session):
        assert not session.has_css("abbr")
        assert not session.has_css("p a#doesnotexist")
        assert not session.has_css("p.nosuchclass")

    def test_respects_scopes(self, session):
        with session.scope("//p[@id='first']"):
            assert session.has_css("a#foo")
            assert not session.has_css("a#red")

    @pytest.mark.requires("js")
    def test_waits_for_content_to_appear(self, session):
        session.visit("/with_js")
        session.click_link("Click me")
        assert session.has_css("input[type='submit'][value='New Here']")

    def test_is_true_if_the_content_occurs_within_the_range_given(self, session):
        assert session.has_css("p", between=range(1, 5))
        assert session.has_css("p a#foo", between=range(1, 4))
        assert session.has_css("p a.doesnotexist", between=range(0, 9))

    def test_is_false_if_the_content_occurs_more_or_fewer_times_than_range(self, session):
        assert not session.has_css("p", between=range(6, 12))
        assert not session.has_css("p a#foo", between=range(4, 8))
        assert not session.has_css("p a.doesnotexist", between=range(3, 9))

    def test_is_true_if_the_content_occurs_the_given_number_of_times(self, session):
        assert session.has_css("p", count=3)
        assert session.has_css("p a#foo", count=1)
        assert session.has_css("p a.doesnotexist", count=0)

    def test_is_false_if_the_content_occurs_a_different_number_of_times_than_given(self, session):
        assert not session.has_css("p", count=6)
        assert not session.has_css("p a#foo", count=2)
        assert not session.has_css("p a.doesnotexist", count=1)

    def test_coerces_count_to_an_integer(self, session):
        assert session.has_css("p", count="3")
        assert session.has_css("p a#foo", count="1")

    def test_is_true_when_content_occurs_same_or_fewer_times_than_given(self, session):
        assert session.has_css("h2.head", maximum=5)
        assert session.has_css("h2", maximum=10)
        assert session.has_css("p a.doesnotexist", maximum=1)
        assert session.has_css("p a.doesnotexist", maximum=0)

    def test_is_false_when_content_occurs_more_times_than_given(self, session):
        assert not session.has_css("h2.head", maximum=4)
        assert not session.has_css("h2", maximum=3)
        assert not session.has_css("p", maximum=1)

    def test_coerces_maximum_to_an_integer(self, session):
        assert session.has_css("h2.head", maximum="5")
        assert session.has_css("h2", maximum="10")

    def test_is_true_when_content_occurs_same_or_more_times_than_given(self, session):
        assert session.has_css("h2.head", mimimum=5)
        assert session.has_css("h2", minimum=3)
        assert session.has_css("p a.doesnotexist", minimum=0)

    def test_is_false_when_content_occurs_fewer_times_than_given(self, session):
        assert not session.has_css("h2.head", minimum=6)
        assert not session.has_css("h2", minimum=8)
        assert not session.has_css("p", minimum=10)
        assert not session.has_css("p a.doesnotexist", minimum=1)

    def test_coerces_minimum_to_an_integer(self, session):
        assert session.has_css("h2.head", minimum="5")
        assert session.has_css("h2", minimum="3")

    def test_discards_all_matches_where_the_given_string_is_not_contained(self, session):
        assert session.has_css("p a", text="Redirect", count=1)
        assert not session.has_css("p a", text="Doesnotexist")

    def test_discards_all_matches_where_the_given_regex_is_not_matched(self, session):
        assert session.has_css("p a", text=re.compile("re[dab]i", re.IGNORECASE), count=1)
        assert not session.has_css("p a", text=re.compile("Red$"))


class TestHasNoCSS:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")

    def test_is_false_if_the_given_selector_is_on_the_page(self, session):
        assert not session.has_no_css("p")
        assert not session.has_no_css("p a#foo")

    def test_is_true_if_the_given_selector_is_not_on_the_page(self, session):
        assert session.has_no_css("abbr")
        assert session.has_no_css("p a#doesnotexist")
        assert session.has_no_css("p.nosuchclass")

    def test_respects_scopes(self, session):
        with session.scope("//p[@id='first']"):
            assert not session.has_no_css("a#foo")
            assert session.has_no_css("a#red")

    @pytest.mark.requires("js")
    def test_waits_for_content_to_disappear(self, session):
        session.visit("/with_js")
        session.click_link("Click me")
        assert session.has_no_css("p#change")

    def test_is_false_if_the_content_occurs_within_the_range_given(self, session):
        assert not session.has_no_css("p", between=range(1, 5))
        assert not session.has_no_css("p a#foo", between=range(1, 4))
        assert not session.has_no_css("p a.doesnotexist", between=range(0, 9))

    def test_is_true_if_the_content_occurs_more_or_fewer_times_than_range(self, session):
        assert session.has_no_css("p", between=range(6, 12))
        assert session.has_no_css("p a#foo", between=range(4, 8))
        assert session.has_no_css("p a.doesnotexist", between=range(3, 9))

    def test_is_false_if_the_content_occurs_the_given_number_of_times(self, session):
        assert not session.has_no_css("p", count=3)
        assert not session.has_no_css("p a#foo", count=1)
        assert not session.has_no_css("p a.doesnotexist", count=0)

    def test_is_true_if_the_content_occurs_a_different_number_of_times_than_given(self, session):
        assert session.has_no_css("p", count=6)
        assert session.has_no_css("p a#foo", count=2)
        assert session.has_no_css("p a.doesnotexist", count=1)

    def test_coerces_count_to_an_integer(self, session):
        assert not session.has_no_css("p", count="3")
        assert not session.has_no_css("p a#foo", count="1")

    def test_is_false_when_content_occurs_same_or_fewer_times_than_given(self, session):
        assert not session.has_no_css("h2.head", maximum=5)
        assert not session.has_no_css("h2", maximum=10)
        assert not session.has_no_css("p a.doesnotexist", maximum=0)

    def test_is_true_when_content_occurs_more_times_than_given(self, session):
        assert session.has_no_css("h2.head", maximum=4)
        assert session.has_no_css("h2", maximum=3)
        assert session.has_no_css("p", maximum=1)

    def test_coerces_maximum_to_an_integer(self, session):
        assert not session.has_no_css("h2.head", maximum="5")
        assert not session.has_no_css("h2", maximum="10")

    def test_is_false_when_content_occurs_same_or_more_times_than_given(self, session):
        assert not session.has_no_css("h2.head", mimimum=5)
        assert not session.has_no_css("h2", minimum=3)
        assert not session.has_no_css("p a.doesnotexist", minimum=0)

    def test_is_true_when_content_occurs_fewer_times_than_given(self, session):
        assert session.has_no_css("h2.head", minimum=6)
        assert session.has_no_css("h2", minimum=8)
        assert session.has_no_css("p", minimum=15)
        assert session.has_no_css("p a.doesnotexist", minimum=1)

    def test_coerces_minimum_to_an_integer(self, session):
        assert not session.has_no_css("h2.head", minimum="5")
        assert not session.has_no_css("h2", minimum="3")

    def test_discards_all_matches_where_the_given_string_is_contained(self, session):
        assert not session.has_no_css("p a", text="Redirect", count=1)
        assert session.has_no_css("p a", text="Doesnotexist")

    def test_discards_all_matches_where_the_given_regex_is_matched(self, session):
        assert not session.has_no_css("p a", text=re.compile("re[dab]i", re.IGNORECASE), count=1)
        assert session.has_no_css("p a", text=re.compile("Red$"))



from xpath import dsl as x

import capybara
from capybara.helpers import desc
from capybara.selector.filter_set import add_filter_set, remove_filter_set, filter_sets
from capybara.selector.selector import add_selector, remove_selector, selectors
from capybara.utils import isregex


__all__ = ["add_filter_set", "add_selector", "filter_sets", "remove_filter_set", "remove_selector",
           "selectors"]


with add_selector("css") as s:
    s.css = lambda css: css

with add_selector("xpath") as s:
    s.xpath = lambda xpath: xpath

with add_selector("id") as s:
    @s.xpath
    def xpath(id):
        return x.descendant()[x.attr("id") == id]

with add_filter_set("field") as fs:
    @fs.filter("checked", boolean=True)
    def checked(node, value):
        return not node.checked ^ value

    @fs.filter("disabled", boolean=True, default=False, skip_if="all")
    def disabled(node, value):
        return not node.disabled ^ value

    @fs.filter("readonly", boolean=True)
    def readonly(node, value):
        return not node.readonly ^ value

    @fs.filter("unchecked", boolean=True)
    def unchecked(node, value):
        return node.checked ^ value

    @fs.describe
    def describe(options):
        description, states = "", []

        if options.get("checked") or options.get("unchecked") is False:
            states.append("checked")
        if options.get("unchecked") or options.get("checked") is False:
            states.append("not checked")
        if options.get("disabled") is True:
            states.append("disabled")

        if states:
            description += " that is {}".format(" and ".join(states))

        return description

with add_selector("button") as s:
    @s.xpath
    def xpath(locator):
        input_button_expr = x.descendant("input")[
            x.attr("type").one_of("submit", "reset", "image", "button")]
        button_expr = x.descendant("button")
        image_button_expr = x.descendant("input")[x.attr("type").equals("image")]

        attr_matchers = (
            x.attr("id").equals(locator) |
            x.attr("value").is_(locator) |
            x.attr("title").is_(locator))
        image_attr_matchers = x.attr("alt").is_(locator)

        if capybara.enable_aria_label:
            attr_matchers |= x.attr("aria-label").is_(locator)
            image_attr_matchers |= x.attr("aria-label").is_(locator)

        input_button_expr = input_button_expr[attr_matchers]
        button_expr = button_expr[
            attr_matchers |
            x.string.n.is_(locator) |
            x.descendant("img")[x.attr("alt").is_(locator)]]
        image_button_expr = image_button_expr[image_attr_matchers]

        return input_button_expr + button_expr + image_button_expr

    @s.filter("disabled", boolean=True, default=False, skip_if="all")
    def disabled(node, value):
        return not node.disabled ^ value

    @s.describe
    def describe(options):
        description = ""
        if options.get("disabled") is True:
            description += " that is disabled"
        return description

with add_selector("checkbox") as s:
    @s.xpath
    def xpath(locator):
        expr = x.descendant("input")[x.attr("type").equals("checkbox")]
        expr = _locate_field(expr, locator)
        return expr

    s.filter_set("field")

with add_selector("field") as s:
    @s.xpath
    def xpath(locator):
        expr = x.descendant("input", "select", "textarea")[
            ~x.attr("type").one_of("hidden", "image", "submit")]
        expr = _locate_field(expr, locator)
        return expr

    s.filter_set("field")

with add_selector("fieldset") as s:
    @s.xpath
    def xpath(locator):
        expr = x.descendant("fieldset")
        expr = expr[
            x.attr("id").equals(locator) |
            x.child("legend")[x.string.n.is_(locator)]]
        return expr

with add_selector("file_field") as s:
    s.label = "file field"

    @s.xpath
    def xpath(locator):
        expr = x.descendant("input")[x.attr("type").equals("file")]
        expr = _locate_field(expr, locator)
        return expr

    s.filter_set("field")

with add_selector("fillable_field") as s:
    s.label = "field"

    @s.xpath
    def xpath(locator):
        expr = x.descendant("input", "textarea")[
            ~x.attr("type").one_of("checkbox", "file", "hidden", "image", "radio", "submit")]
        expr = _locate_field(expr, locator)
        return expr

    s.filter_set("field")

with add_selector("frame") as s:
    @s.xpath
    def xpath(locator):
        expr = x.descendant("frame") + x.descendant("iframe")
        expr = expr[x.attr("id").equals(locator) | x.attr("name").equals(locator)]
        return expr

with add_selector("label") as s:
    @s.xpath
    def xpath(locator):
        expr = x.descendant("label")
        if locator:
            expr = expr[x.string.n.is_(str(locator)) | x.attr("id").equals(str(locator))]
        return expr

    @s.filter("field")
    def field(node, field_or_value):
        from capybara.node.element import Element

        if isinstance(field_or_value, Element):
            if field_or_value["id"] and field_or_value["id"] == node["for"]:
                return True
            else:
                return node.base in field_or_value._find_xpath("./ancestor::label[1]")
        else:
            return node["for"] == str(field_or_value)

    @s.describe
    def describe(options):
        description = ""
        if options.get("field"):
            description += " for {}".format(options["field"])
        return description

with add_selector("link") as s:
    @s.xpath
    def xpath(locator):
        expr = x.descendant("a")[x.attr("href")]

        attr_matchers = (
            x.attr("id").equals(locator) |
            x.attr("title").is_(locator) |
            x.string.n.is_(locator))

        if capybara.enable_aria_label:
            attr_matchers |= x.attr("aria-label").is_(locator)

        expr = expr[
            attr_matchers |
            x.descendant("img")[x.attr("alt").is_(locator)]]

        return expr

    @s.filter("href")
    def href(node, href):
        if isregex(href):
            return bool(href.search(node["href"]))
        else:
            # For href element attributes, Selenium returns the full URL that would
            # be visited rather than the raw value in the source. So we use XPath.
            query = x.axis("self")[x.attr("href") == str(href)]
            return node.has_selector("xpath", query)

    @s.describe
    def describe(options):
        description = ""
        if options.get("href"):
            description += " with href {}".format(desc(options["href"]))
        return description

with add_selector("link_or_button") as s:
    s.label = "link or button"

    @s.xpath
    def xpath(locator):
        return selectors["link"](locator) + selectors["button"](locator)

    @s.filter("disabled", boolean=True, default=False, skip_if="all")
    def disabled(node, value):
        return (
            node.tag_name == "a" or
            not node.disabled ^ value)

    @s.describe
    def describe(options):
        description = ""
        if options.get("disabled") is True:
            description += " that is disabled"
        return description

with add_selector("option") as s:
    @s.xpath
    def xpath(locator):
        expr = x.descendant("option")
        expr = expr[x.string.n.is_(locator)]
        return expr

with add_selector("radio_button") as s:
    s.label = "radio button"

    @s.xpath
    def xpath(locator):
        expr = x.descendant("input")[x.attr("type").equals("radio")]
        expr = _locate_field(expr, locator)
        return expr

    s.filter_set("field")

with add_selector("select") as s:
    s.label = "select box"

    @s.xpath
    def xpath(locator):
        expr = x.descendant("select")
        expr = _locate_field(expr, locator)
        return expr

    s.filter_set("field")

    @s.filter("multiple", boolean=True)
    def multiple(node, value):
        return not node.multiple ^ value

    @s.filter("options")
    def options(node, options):
        if node.visible:
            actual = [n.text for n in node.find_all("xpath", ".//option")]
        else:
            actual = [n.all_text for n in node.find_all("xpath", ".//option", visible=False)]

        return sorted(options) == sorted(actual)

    @s.filter("selected")
    def selected(node, selected):
        if not isinstance(selected, list):
            selected = [selected]

        actual = [
            n.all_text
            for n in node.find_all("xpath", ".//option", visible=False)
            if n.selected]

        return sorted(selected) == sorted(actual)

    @s.filter("with_options")
    def with_options(node, options):
        finder_settings = {"minimum": 0}
        if not node.visible:
            finder_settings["visible"] = False

        return all([
            node.find_first("option", option, **finder_settings)
            for option in options])

    @s.describe
    def describe(options):
        description = ""

        if options.get("multiple") is True:
            description += " with the multiple attribute"
        if options.get("multiple") is False:
            description += " without the multiple attribute"
        if options.get("options"):
            description += " with options {}".format(desc(options["options"]))
        if options.get("selected"):
            description += " with {} selected".format(desc(options["selected"]))
        if options.get("with_options"):
            description += " with at least options {}".format(desc(options["with_options"]))

        return description

with add_selector("table") as s:
    @s.xpath
    def xpath(locator):
        expr = x.descendant("table")
        expr = expr[
            x.attr("id").equals(locator) |
            x.descendant("caption").is_(locator)]
        return expr


def _locate_field(field_expr, locator):
    attr_matchers = (
        x.attr("id").equals(locator) |
        x.attr("name").equals(locator) |
        x.attr("placeholder").equals(locator) |
        x.attr("id").equals(x.anywhere("label")[x.string.n.is_(locator)].attr("for")))

    if capybara.enable_aria_label:
        attr_matchers |= x.attr("aria-label").is_(locator)

    expr = field_expr[attr_matchers]
    expr += x.descendant("label")[x.string.n.is_(locator)].descendant(field_expr)

    return expr



import pytest

import capybara
from capybara.exceptions import ElementNotFound
from capybara.tests.helpers import extract_results


class UncheckTestCase:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")


class TestUncheck(UncheckTestCase):
    def test_unchecks_a_checkbox_by_id(self, session):
        session.uncheck("form_pets_hamster")
        session.click_button("awesome")
        pets = extract_results(session).getlist("form[pets][]")
        assert "dog" in pets
        assert "hamster" not in pets

    def test_unchecks_a_checkbox_by_label(self, session):
        session.uncheck("Hamster")
        session.click_button("awesome")
        pets = extract_results(session).getlist("form[pets][]")
        assert "dog" in pets
        assert "hamster" not in pets


class TestUncheckWithAutomaticLabelClick(UncheckTestCase):
    @pytest.fixture(autouse=True)
    def setup_settings(self):
        old_automatic_label_click = capybara.automatic_label_click
        capybara.automatic_label_click = True
        try:
            yield
        finally:
            capybara.automatic_label_click = old_automatic_label_click

    def test_unchecks_via_clicking_the_label_with_for_attribute_if_possible(self, session):
        assert session.find("checkbox", "form_cars_jaguar", checked=True, visible="hidden")
        session.uncheck("form_cars_jaguar")
        session.click_button("awesome")
        assert "jaguar" not in extract_results(session).getlist("form[cars][]")

    def test_unchecks_via_clicking_the_wrapping_label_if_possible(self, session):
        assert session.find("checkbox", "form_cars_koenigsegg", checked=True, visible="hidden")
        session.uncheck("form_cars_koenigsegg")
        session.click_button("awesome")
        assert "koenigsegg" not in extract_results(session).getlist("form[cars][]")

    def test_does_not_click_the_label_if_unneeded(self, session):
        assert session.find("checkbox", "form_cars_tesla", unchecked=True, visible="hidden")
        session.uncheck("form_cars_tesla")
        session.click_button("awesome")
        assert "tesla" not in extract_results(session).getlist("form[cars][]")

    def test_raises_original_error_when_no_label_available(self, session):
        with pytest.raises(ElementNotFound) as excinfo:
            session.uncheck("form_cars_ariel")
        assert "Unable to find checkbox 'form_cars_ariel'" in str(excinfo.value)

    def test_raises_error_if_not_allowed_to_click_label(self, session):
        with pytest.raises(ElementNotFound) as excinfo:
            session.uncheck("form_cars_jaguar", allow_label_click=False)
        assert "Unable to find checkbox 'form_cars_jaguar'" in str(excinfo.value)



from capybara.exceptions import CapybaraError


def assert_windows_open(session, count):
    @session.document.synchronize(wait=2, errors=(CapybaraError,))
    def assert_windows_open():
        if len(session.windows) != count:
            raise CapybaraError()
    assert_windows_open()



from warnings import warn

import capybara
from capybara.node.base import Base, synchronize


class Element(Base):
    """
    An :class:`Element` represents a single element on the page. It is possible to interact with the
    contents of this element the same as with a document::

        session = Session("selenium", my_app)

        bar = session.find("#bar")       # from capybara.node.finders.FindersMixin
        bar.select("Baz", field="Quox")  # from capybara.node.actions.ActionsMixin

    :class:`Element` also has access to HTML attributes and other properties of the element::

        bar.value
        bar.text
        bar["title"]
    """

    def __init__(self, session, base, query_scope, query):
        super(type(self), self).__init__(session, base)
        self.query_scope = query_scope
        self.query = query

    def __repr__(self):
        return "<capybara.node.element.Element tag=\"{tag}\" path=\"{path}\">".format(
            tag=self.tag_name,
            path=self.path)

    @property
    def native(self):
        """ object: The native element from the driver. """
        return self.base.native

    def reload(self):
        if self.allow_reload:
            query_scope = self.query_scope.reload()
            reloaded = query_scope.find_first(
                self.query.name, self.query.locator, **self.query.kwargs)
            if reloaded:
                self.base = reloaded.base
        return self

    @property
    @synchronize
    def tag_name(self):
        """ str: The tag name of the element. """
        return self.base.tag_name

    @property
    @synchronize
    def visible(self):
        """ bool: Whether or not the element is visible. """
        return self.base.visible

    @property
    @synchronize
    def value(self):
        """ str: The value of the form element. """
        return self.base.value

    @property
    def text(self):
        """
        Retrieve the text of the element. If :data:`capybara.ignore_hidden_elements` is ``True``,
        which it is by default, then this will return only text which is visible. The exact
        semantics of this may differ between drivers, but generally any text within elements with
        ``display: none`` is ignored.

        Returns:
            str: The text of the element.
        """

        if capybara.ignore_hidden_elements or capybara.visible_text_only:
            return self.visible_text
        else:
            return self.all_text

    @property
    @synchronize
    def all_text(self):
        """ str: All of the text of the element. """
        return self.base.all_text

    @property
    @synchronize
    def visible_text(self):
        """ str: Only the visible text of the element. """
        return self.base.visible_text

    @property
    @synchronize
    def checked(self):
        """ bool: Whether or not the element is checked. """
        return self.base.checked

    @property
    @synchronize
    def selected(self):
        """ bool: Whether or not the element is selected. """
        return self.base.selected

    @property
    @synchronize
    def disabled(self):
        """ bool: Whether or not the element is disabled. """
        return self.base.disabled

    @property
    @synchronize
    def readonly(self):
        """ bool: Whether or not the element is readonly. """
        return self.base.readonly

    @property
    @synchronize
    def multiple(self):
        """ bool: Whether or not the element supports multiple results. """
        return self.base.multiple

    @property
    @synchronize
    def path(self):
        """ str: An XPath expression describing where on the page the element can be found. """
        return self.base.path

    @synchronize
    def __getitem__(self, name):
        """
        Retrieve the given attribute. ::

            element["title"]  # => HTML title attribute

        Args:
            name (str): The attribute to retrieve.

        Returns:
            str: The value of the attribute.
        """

        return self.base[name]

    @synchronize
    def click(self):
        """ Click the element. """
        self.base.click()

    @synchronize
    def double_click(self):
        """ Double-click the element. """
        self.base.double_click()

    @synchronize
    def drag_to(self, node):
        """
        Drag the element to the given other element. ::

            source = page.find("#foo")
            target = page.find("#bar")
            source.drag_to(target)

        Args:
            node (Element): The element to drag to.
        """

        self.base.drag_to(node.base)

    @synchronize
    def send_keys(self, *args):
        """
        Send keystrokes to the element.

        Examples::

            from selenium.webdriver.common.keys import Keys

            element.send_keys("foo")                  # => value: "foo"
            element.send_keys("tet", Keys.LEFT, "s")  # => value: "test"

        Args:
            *args: Variable length list of keys to send.
        """

        self.base.send_keys(*args)

    @synchronize
    def hover(self):
        """ Hover on the element. """
        self.base.hover()

    @synchronize
    def right_click(self):
        """ Right-click the element. """
        self.base.right_click()

    @synchronize
    def select_option(self):
        """ Select this node if it is an option element inside a select tag. """
        if self.disabled:
            warn("Attempt to select disabled option: {}".format(self.value or self.text))
        self.base.select_option()

    @synchronize
    def set(self, value):
        """
        Set the value of the form element to the given value.

        Args:
            value (bool | str): The new value.
        """

        self.base.set(value)

    @synchronize
    def unselect_option(self):
        """ Unselect this node if it is an option element inside a multiple select tag. """
        self.base.unselect_option()



import pytest


@pytest.mark.requires("js")
class TestExecuteScript:
    def test_executes_the_given_script_and_returns_nothing(self, session):
        session.visit("/with_js")
        assert session.execute_script("document.getElementById('change').textContent = 'Funky Doodle'") is None
        assert session.has_css("#change", text="Funky Doodle")

    def test_able_to_call_functions_defined_in_the_page(self, session):
        session.visit("/with_js")
        session.execute_script("$('#change').text('Funky Doodle')")

    def test_passes_arguments_to_the_script(self, session):
        session.visit("/with_js")
        session.execute_script("document.getElementById('change').textContent = arguments[0]", "Doodle Funk")
        assert session.has_css("#change", text="Doodle Funk")

    def test_supports_passing_elements_as_arguments_to_the_script(self, session):
        session.visit("/with_js")
        el = session.find("css", "#change")
        session.execute_script("arguments[1].textContent = arguments[0]", "Doodle Funk", el)
        assert session.has_css("#change", text="Doodle Funk")



import flask
from flask import Flask, render_template, request
import json
import time


app = Flask(__name__)


class TestAppError(Exception):
    pass


@app.route("/")
def hello():
    resp = flask.make_response("""Hello world! <a href="with_html">Relative</a>""")
    resp.set_cookie("capybara", "root cookie")
    return resp


@app.route("/foo")
def foo():
    return "Another World"


@app.route("/redirect", methods=["GET", "POST"])
def redirect():
    return flask.redirect("/redirect_again")


@app.route("/redirect_again")
def redirect_again():
    return flask.redirect("/landed")


@app.route("/referer_base")
def referer_base():
    return """
    <a href="/get_referrer">direct link</a>
    <a href="/redirect_to_get_referrer">link via redirect</a>
    <form action="/get_referrer" method="get"><input type="submit"></form>
    """


@app.route("/redirect_to_get_referrer")
def redirect_to_get_referrer():
    return flask.redirect("/get_referrer")


@app.route("/get_referrer")
def get_referrer():
    return "No referer" if request.referrer is None else "Got referer: {0}".format(request.referrer)


@app.route("/host")
def host():
    return "Current host is {0}://{1}".format(request.scheme, request.host)


@app.route("/redirect/<int:times>/times")
def redirect_n_times(times):
    if times == 0:
        return "redirection complete"
    else:
        return flask.redirect("/redirect/{0}/times".format(times - 1))


@app.route("/landed")
def landed():
    return "You landed"


@app.route("/with-quotes")
def with_quotes():
    return "\"No,\" he said, \"you can't do that.\""


@app.route("/form/get", methods=["GET"])
@app.route("/relative", methods=["POST"])
@app.route("/form", methods=["POST"])
def results():
    data = request.args.copy()
    data.update(request.form)
    return """<pre id="results">""" + json.dumps(data.to_dict(flat=False)) + """</pre>"""


@app.route("/favicon.ico")
def favicon():
    return ""


@app.route("/delete", methods=["DELETE"])
def deleted():
    return "The requested object was deleted"


@app.route("/delete", methods=["GET"])
def not_deleted():
    return "Not deleted"


@app.route("/redirect_back")
def redirect_back():
    return flask.redirect(request.referrer)


@app.route("/redirect_secure")
def redirect_secure():
    return flask.redirect("http://{0}/host".format(request.host))


@app.route("/slow_response")
def slow_response():
    time.sleep(2)
    return "Finally!"


@app.route("/set_cookie")
def set_cookie():
    cookie_value = "test_cookie"
    resp = flask.make_response("Cookie set to {0}".format(cookie_value))
    resp.set_cookie("capybara", cookie_value)
    return resp


@app.route("/get_cookie")
def get_cookie():
    return request.cookies.get("capybara", "")


@app.route("/get_header")
def get_header():
    return request.headers.get("Foo", "")


@app.route("/get_header_via_redirect")
def get_header_via_redirect():
    return flask.redirect("/get_header")


@app.route("/error")
def error():
    raise TestAppError()


@app.route("/import_error")
def import_error():
    raise ImportError("Simulated ImportError")


@app.route("/with_html")
def with_html():
    return render_template("with_html.html")


@app.route("/with_simple_html")
def with_simple_html():
    return render_template("with_simple_html.html")


@app.route("/<name>")
def view(name):
    return render_template("{}.html".format(name))


@app.route("/upload_empty", methods=["POST"])
def upload_empty():
    f = request.files.get("form[file]")

    if not f:
        return "Successfully ignored empty file field."
    else:
        return "Something went wrong."


@app.route("/upload", methods=["POST"])
def upload():
    document = request.files.get("form[document]")

    if document and document.filename:
        buf = []
        buf.append("Content-type: {0}".format(document.mimetype))
        buf.append("File content: {0}".format(document.read()))
        return " | ".join(buf)
    else:
        return "No file uploaded"


@app.route("/upload_multiple", methods=["POST"])
def upload_multiple():
    documents = request.files.getlist("form[multiple_documents][]")
    documents = [doc for doc in documents if doc.filename]

    if len(documents):
        buf = [str(len(documents))]

        for document in documents:
            buf.append("Content-type: {0}".format(document.mimetype))
            buf.append("File content: {0}".format(document.read()))

        return " | ".join(buf)
    else:
        return "No files uploaded"


if __name__ == "__main__":
    app.run()



import pytest

import capybara
from capybara.exceptions import ElementNotFound


class FindButtonTestCase:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")


class TestFindButton(FindButtonTestCase):
    def test_finds_any_button(self, session):
        assert session.find_button("med")["id"] == "mediocre"
        assert session.find_button("crap321").value == "crappy"

    def test_raises_an_error_if_the_button_does_not_exist(self, session):
        with pytest.raises(ElementNotFound):
            session.find_button("Does not exist")

    def test_raises_an_error_if_the_button_is_disabled(self, session):
        with pytest.raises(ElementNotFound):
            session.find_button("Disabled button")

    def test_finds_by_aria_label_attribute_when_enable_aria_label_is_true(self, session):
        capybara.enable_aria_label = True
        assert session.find_button("Mediocre Button")["id"] == "mediocre"

    def test_does_not_find_by_aria_label_attribute_when_enable_aria_label_is_false(self, session):
        capybara.enable_aria_label = False
        with pytest.raises(ElementNotFound):
            session.find_button("Mediocre Button")


class TestFindButtonDisabled(FindButtonTestCase):
    def test_does_not_find_disabled_buttons_when_false(self, session):
        with pytest.raises(ElementNotFound):
            session.find_button("Disabled button", disabled=False)

    def test_finds_enabled_buttons_when_false(self, session):
        assert session.find_button("med", disabled=False)["id"] == "mediocre"

    def test_finds_disabled_buttons_when_true(self, session):
        field = session.find_button("Disabled button", disabled=True)
        assert field.value == "Disabled button"

    def test_does_not_find_enabled_buttons_when_true(self, session):
        with pytest.raises(ElementNotFound):
            session.find_button("med", disabled=True)

    def test_finds_disabled_buttons_when_all(self, session):
        field = session.find_button("Disabled button", disabled="all")
        assert field.value == "Disabled button"

    def test_finds_enabled_buttons_when_all(self, session):
        assert session.find_button("med", disabled="all")["id"] == "mediocre"



from selenium import webdriver
from selenium.webdriver import DesiredCapabilities


def get_browser(browser_name, capabilities=None):
    """
    Returns an instance of the given browser with the given capabilities.

    Args:
        browser_name (str): The name of the desired browser.
        capabilities (Dict[str, str | bool], optional): The desired capabilities of the browser. Defaults to None.

    Returns:
        WebDriver: An instance of the desired browser.
    """

    if browser_name == "chrome":
        return webdriver.Chrome(desired_capabilities=capabilities)
    if browser_name == "edge":
        return webdriver.Edge(capabilities=capabilities)
    if browser_name in ["ff", "firefox"]:
        return webdriver.Firefox(capabilities=capabilities)
    if browser_name in ["ie", "internet_explorer"]:
        return webdriver.Ie(capabilities=capabilities)
    if browser_name == "phantomjs":
        return webdriver.PhantomJS(desired_capabilities=capabilities)
    if browser_name == "safari":
        return webdriver.Safari(desired_capabilities=capabilities)

    raise ValueError("unsupported browser: {}".format(repr(browser_name)))



import pytest

from capybara.exceptions import ElementNotFound


class TestFindByID:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")

    def test_finds_any_element_by_id(self, session):
        assert session.find_by_id("red").tag_name == "a"

    def test_raises_an_error_if_no_element_with_id_is_found(self, session):
        with pytest.raises(ElementNotFound):
            session.find_by_id("nothing_with_this_id")

    def test_finds_invisible_elements_when_visible_is_false(self, session):
        assert "with hidden ancestor" in session.find_by_id("hidden_via_ancestor", visible=False).all_text

    def test_does_not_find_invisible_elements_when_visible_is_true(self, session):
        with pytest.raises(ElementNotFound):
            session.find_by_id("hidden_via_ancestor", visible=True)



from __future__ import absolute_import

import pytest

from capybara.tests.app import app


class DriverSuite(object):
    """
    Represents a suite of tests that should be run against a particular driver.

    Args:
        driver_name (str): The registered name of the driver to test.
        skip (List[str], optional): A list of features not supported by the driver. Tests in
            the suite marked as requiring these features will be skipped.
    """

    def __init__(self, driver_name, skip=None):
        self.driver_name = driver_name
        self.skip = skip or []

    @pytest.fixture(scope="session")
    def session(self):
        from capybara.session import Session

        return Session(self.driver_name, app)

    @pytest.fixture(autouse=True)
    def reset_session(self, session):
        try:
            yield
        finally:
            session.reset()



import pytest
from xpath import html

import capybara


class FindFirstTestCase:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")


class TestFindFirst(FindFirstTestCase):
    def test_finds_the_first_element_using_the_given_locator(self, session):
        assert session.find_first("//h1").text == "This is a test"
        assert session.find_first("//input[@id='test_field']").value == "monkey"

    def test_returns_none_when_nothing_was_found(self, session):
        assert session.find_first("//div[@id='nosuchthing']") is None

    def test_accepts_an_xpath_expression_instance(self, session):
        session.visit("/form")
        xpath = html.fillable_field("First Name")
        assert session.find_first(xpath).value == "John"

    def test_finds_the_first_element_using_the_given_css_selector(self, session):
        assert session.find_first("css", "h1").text == "This is a test"
        assert session.find_first("css", "input[id='test_field']").value == "monkey"

    def test_finds_the_first_element_using_the_given_xpath_query(self, session):
        assert session.find_first("xpath", "//h1").text == "This is a test"
        assert session.find_first("xpath", "//input[@id='test_field']").value == "monkey"

    def test_finds_the_first_element_using_the_given_locator_when_css_is_the_default_selector(self, session):
        capybara.default_selector = "css"
        assert session.find_first("h1").text == "This is a test"
        assert session.find_first("input[id='test_field']").value == "monkey"

    def test_finds_the_first_element_using_the_given_locator_in_a_scope(self, session):
        session.visit("/with_scope")
        with session.scope("xpath", "//div[@id='for_bar']"):
            assert session.find_first(".//form") is not None


class TestFindFirstVisible(FindFirstTestCase):
    def test_only_finds_visible_nodes_when_true(self, session):
        assert session.find_first("css", "a#invisible", visible=True) is None
        assert session.find_first("css", "a#visible", visible=True) is not None

    def test_finds_nodes_regardless_of_whether_they_are_invisible_when_false(self, session):
        assert session.find_first("css", "a#invisible", visible=False) is not None
        assert session.find_first("css", "a#visible", visible=False) is not None

    def test_finds_nodes_regardless_of_whether_they_are_invisible_when_all(self, session):
        assert session.find_first("css", "a#invisible", visible="all") is not None
        assert session.find_first("css", "a#visible", visible="all") is not None

    def test_finds_only_hidden_nodes_when_hidden(self, session):
        assert session.find_first("css", "a#invisible", visible="hidden") is not None
        assert session.find_first("css", "a#visible", visible="hidden") is None

    def test_finds_only_visible_nodes_when_visible(self, session):
        assert session.find_first("css", "a#invisible", visible="visible") is None
        assert session.find_first("css", "a#visible", visible="visible") is not None


@pytest.mark.requires("js")
class TestFindFirstWithWaitOnFirstByDefault(FindFirstTestCase):
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_js")

    def test_does_not_wait_if_false(self, session):
        capybara.wait_on_first_by_default = False
        session.click_link("clickable")
        assert session.find_first("css", "a#has-been-clicked") is None

    def test_waits_for_at_least_one_match_if_true(self, session):
        capybara.wait_on_first_by_default = True
        session.click_link("clickable")
        assert session.find_first("css", "a#has-been-clicked") is not None

    def test_returns_nil_after_waiting_if_no_match(self, session):
        capybara.wait_on_first_by_default = True
        session.click_link("clickable")
        assert session.find_first("css", "a#not-a-real-link") is None



import pytest
import re

from capybara.exceptions import ElementNotFound


class TestClickLink:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")

    def test_follows_links_by_id(self, session):
        session.click_link("foo")
        assert session.has_text("Another World")

    def test_follows_links_by_text(self, session):
        session.click_link("labore")
        assert session.has_text("Bar")

    def test_follows_links_by_partial_text(self, session):
        session.click_link("abo")
        assert session.has_text("Bar")

    def test_follows_links_by_title(self, session):
        session.click_link("awesome title")
        assert session.has_text("Bar")

    def test_follows_links_by_partial_title(self, session):
        session.click_link("some titl")
        assert session.has_text("Bar")

    def test_follows_links_by_alternate_text_of_descendant_image(self, session):
        session.click_link("awesome image")
        assert session.has_text("Bar")

    def test_follows_links_by_partial_alternate_text_of_descendant_image(self, session):
        session.click_link("some imag")
        assert session.has_text("Bar")

    def test_raises_an_error_for_a_locator_that_does_not_exist(self, session):
        with pytest.raises(ElementNotFound):
            session.click_link("does not exist")

    def test_finds_links_with_a_given_href(self, session):
        session.click_link("labore", href="/with_simple_html")
        assert session.has_text("Bar")

    def test_raises_if_link_with_given_href_not_found(self, session):
        with pytest.raises(ElementNotFound):
            session.click_link("labore", href="invalid_href")

    def test_finds_a_link_matching_an_all_matching_regex_pattern(self, session):
        session.click_link("labore", href=re.compile(r".+"))
        assert session.has_text("Bar")

    def test_finds_a_link_matching_an_exact_regex_pattern(self, session):
        session.click_link("labore", href=re.compile(r"\/with_simple_html"))
        assert session.has_text("Bar")

    def test_finds_a_link_matching_a_partial_regex_pattern(self, session):
        session.click_link("labore", href=re.compile(r"\/with_simple"))
        assert session.has_text("Bar")

    def test_raises_an_error_if_no_link_href_matched_the_pattern(self, session):
        with pytest.raises(ElementNotFound):
            session.click_link("labore", href=re.compile(r"invalid_pattern"))
        with pytest.raises(ElementNotFound):
            session.click_link("labore", href=re.compile(r".+d+"))

    def test_does_nothing_on_anchor_links(self, session):
        session.fill_in("test_field", value="blah")
        session.click_link("Normal Anchor")
        assert session.find_field("test_field").value == "blah"
        session.click_link("Blank Anchor")
        assert session.find_field("test_field").value == "blah"
        session.click_link("Blank JS Anchor")
        assert session.find_field("test_field").value == "blah"

    def test_raises_an_error_for_links_with_no_href(self, session):
        with pytest.raises(ElementNotFound):
            session.click_link("No Href")

    def test_follows_links_by_partial_match_when_exact_is_false(self, session):
        session.click_link("abo", exact=False)
        assert session.has_text("Bar")

    def test_raises_an_error_for_partial_match_when_exact_is_true(self, session):
        with pytest.raises(ElementNotFound):
            session.click_link("abo", exact=True)



from itertools import chain
from xpath import dsl as x
from xpath.renderer import to_xpath

from capybara.driver.node import Node as Base
from capybara.exceptions import ReadOnlyElementError, UnselectNotAllowed
from capybara.helpers import normalize_whitespace
from capybara.node.simple import Simple
from capybara.utils import inner_text


class Node(Base):
    @property
    def tag_name(self):
        return self.native.tag

    def __getitem__(self, name):
        return self._string_node[name]

    def __eq__(self, other):
        return self.native == other.native

    def __hash__(self):
        return hash(self.native)

    @property
    def all_text(self):
        return normalize_whitespace(inner_text(self.native))

    @property
    def visible_text(self):
        return normalize_whitespace(self.unnormalized_text())

    def unnormalized_text(self, check_ancestor_visibility=True):
        if not self._string_node._visible(check_ancestor_visibility):
            return ""
        else:
            parts = (
                [self.native.text] +
                list(chain(*(
                    [Node(self.driver, c).unnormalized_text(check_ancestor_visibility=False),
                     c.tail]
                    for c in self.native.getchildren()))))
            return "".join(filter(None, parts))

    @property
    def disabled(self):
        if self.tag_name in ["option", "optgroup"]:
            return self._string_node.disabled or self._find_xpath("parent::*")[0].disabled
        else:
            return self._string_node.disabled

    @property
    def readonly(self):
        return self._string_node.readonly

    @property
    def multiple(self):
        return self._string_node.multiple

    @property
    def path(self):
        return self._string_node.path

    @property
    def checked(self):
        return self._string_node.checked

    @property
    def selected(self):
        return self._string_node.selected

    @property
    def visible(self):
        return self._string_node.visible

    @property
    def value(self):
        return self._string_node.value

    def click(self):
        if self.tag_name == "a" and self["href"]:
            self.driver.follow("GET", self["href"])
        elif (self.tag_name == "input" and self["type"] in ["submit", "image"]) or \
                (self.tag_name == "button"):
            associated_form = self._form
            if associated_form is not None:
                from capybara.werkzeug.form import Form
                Form(self.driver, associated_form).submit(self)
        elif self.tag_name == "input" and self["type"] in ["checkbox", "radio"]:
            self.set(not self.checked)
        elif self.tag_name == "label":
            labeled_controls = (
                self._find_xpath("//input[@id='{}']".format(self["for"])) if self["for"]
                else self._find_xpath(".//input"))
            labeled_control = labeled_controls[0] if len(labeled_controls) else None

            if labeled_control and (labeled_control._is_checkbox or labeled_control._is_radio):
                labeled_control.set(not labeled_control.checked)

    def set(self, value):
        if self.readonly:
            raise ReadOnlyElementError

        if self._is_radio:
            self._set_radio(value)
        elif self._is_checkbox:
            self._set_checkbox(value)
        elif self._is_input_field:
            self._set_input(value)
        elif self._is_textarea:
            self._set_textarea(value)

    def select_option(self):
        if self.disabled:
            return

        select = self.native.xpath("ancestor::select")[0]

        if select.get("multiple", None) != "multiple":
            options = select.xpath(".//option[@selected='selected']")
            for option in options:
                option.attrib.pop("selected", None)

        self.native.set("selected", "selected")

    def unselect_option(self):
        select = self.native.xpath("ancestor::select")[0]
        if select.get("multiple", None) != "multiple":
            raise UnselectNotAllowed()

        self.native.attrib.pop("selected", None)

    @property
    def _is_input_field(self):
        return self.tag_name == "input"

    @property
    def _is_checkbox(self):
        return self._is_input_field and self["type"] == "checkbox"

    @property
    def _is_radio(self):
        return self._is_input_field and self["type"] == "radio"

    @property
    def _is_textarea(self):
        return self.tag_name == "textarea"

    @property
    def _string_node(self):
        return Simple(self.native)

    @property
    def _form(self):
        elements = (
            self.native.xpath("//form[@id='{}']".format(self["form"])) if self["form"]
            else self.native.xpath(".//ancestor::form"))

        return elements[0] if elements else None

    def _find_css(self, css):
        return self._find_xpath(to_xpath(x.css(css)))

    def _find_xpath(self, xpath):
        cls = type(self)
        elements = self.native.xpath(xpath)
        return [cls(self.driver, element) for element in elements]

    def _set_radio(self, value):
        other_radios = self.native.xpath(to_xpath(
            x.anywhere("input")[x.attr("name") == self["name"]]))
        for node in other_radios:
            node.attrib.pop("checked", None)

        self.native.set("checked", "checked")

    def _set_checkbox(self, value):
        if value:
            self.native.set("checked", "checked")
        else:
            self.native.attrib.pop("checked", None)

    def _set_input(self, value):
        if self["maxlength"]:
            maxlength = int(self["maxlength"])
            value = value[0:maxlength]

        self.native.set("value", value)

    def _set_textarea(self, value):
        for child in self.native.getchildren():
            self.native.remove(child)

        self.native.text = value



import os
from setuptools import setup, find_packages
import sys


tests_require = ["selenium < 4", "werkzeug"]
if sys.version_info < (3, 3):
    tests_require.append("mock")

driver_verification_tests_require = ["flaky", "flask", "py", "pytest >= 3", "werkzeug"]
if sys.version_info < (3, 3):
    driver_verification_tests_require.append("mock")

tests_require += driver_verification_tests_require


def read(filename):
    """
    Returns the contents of the given package file.

    Args:
        filename (str): The name of the file to read, relative to the current
            directory.

    Returns:
        str: The contents of the given package file.
    """

    path = os.path.join(os.path.dirname(__file__), filename)

    with open(path) as f:
        return f.read()


def get_version():
    """ str: The package version. """

    global_vars = {}

    # Compile and execute the individual file to prevent
    # the package from being automatically loaded.
    source = read(os.path.join("capybara", "version.py"))
    code = compile(source, "version.py", "exec")
    exec(code, global_vars)

    return global_vars['__version__']


setup(
    name="capybara-py",
    version=get_version(),
    description="Acceptance test framework for web applications",
    long_description=read("README.rst"),
    url="https://github.com/elliterate/capybara.py",
    author="Ian Lesperance",
    author_email="ian@elliterate.com",
    license="MIT",
    keywords="capybara",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing"],
    packages=find_packages(exclude=["tests", "tests.*"]),
    install_requires=["lxml", "xpath-py >= 0.0.4"],
    setup_requires=["pytest-runner"],
    tests_require=tests_require,
    extras_require={
        # Expose test dependencies for external scripts, like pip.
        "test": tests_require})
